# AUTOGENERATED ON 2020-01-30 22:15:23.661530
## DO NOT MODIFY THIS FILE MANUALLY

from numpy import inf, nan

from autogoal.grammar import Continuous, Discrete, Categorical, Boolean
from autogoal.contrib.sklearn._builder import SklearnEstimator, SklearnTransformer
from autogoal.kb import *
from autogoal.utils import nice_repr

from sklearn.cluster._affinity_propagation import (
    AffinityPropagation as _AffinityPropagation,
)


@nice_repr
class AffinityPropagation(_AffinityPropagation, SklearnEstimator):
    def __init__(
        self,
        convergence_iter: Discrete(min=1, max=29),
        affinity: Categorical("euclidean"),
    ):
        SklearnEstimator.__init__(self)
        _AffinityPropagation.__init__(
            self, convergence_iter=convergence_iter, affinity=affinity
        )

    def run(
        self, input: Tuple(MatrixContinuousDense(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.cluster._agglomerative import FeatureAgglomeration as _FeatureAgglomeration


@nice_repr
class FeatureAgglomeration(_FeatureAgglomeration, SklearnTransformer):
    def __init__(
        self,
        n_clusters: Discrete(min=1, max=3),
        affinity: Categorical("euclidean"),
        compute_full_tree: Categorical("auto"),
        linkage: Categorical("average", "complete", "single", "ward"),
    ):
        SklearnTransformer.__init__(self)
        _FeatureAgglomeration.__init__(
            self,
            n_clusters=n_clusters,
            affinity=affinity,
            compute_full_tree=compute_full_tree,
            linkage=linkage,
        )

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.cluster._birch import Birch as _Birch


@nice_repr
class Birch(_Birch, SklearnEstimator):
    def __init__(
        self,
        threshold: Continuous(min=-4.995, max=4.991),
        branching_factor: Discrete(min=2, max=99),
        n_clusters: Discrete(min=1, max=5),
        compute_labels: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _Birch.__init__(
            self,
            threshold=threshold,
            branching_factor=branching_factor,
            n_clusters=n_clusters,
            compute_labels=compute_labels,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.cluster._kmeans import KMeans as _KMeans


@nice_repr
class KMeans(_KMeans, SklearnEstimator):
    def __init__(
        self,
        n_clusters: Discrete(min=1, max=15),
        init: Categorical("random"),
        precompute_distances: Categorical("auto"),
    ):
        SklearnEstimator.__init__(self)
        _KMeans.__init__(
            self,
            n_clusters=n_clusters,
            init=init,
            precompute_distances=precompute_distances,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.cluster._kmeans import MiniBatchKMeans as _MiniBatchKMeans


@nice_repr
class MiniBatchKMeans(_MiniBatchKMeans, SklearnEstimator):
    def __init__(
        self,
        n_clusters: Discrete(min=1, max=15),
        init: Categorical("random"),
        compute_labels: Boolean(),
        tol: Continuous(min=-0.992, max=0.992),
        max_no_improvement: Discrete(min=1, max=19),
        reassignment_ratio: Continuous(min=-0.093, max=0.094),
    ):
        SklearnEstimator.__init__(self)
        _MiniBatchKMeans.__init__(
            self,
            n_clusters=n_clusters,
            init=init,
            compute_labels=compute_labels,
            tol=tol,
            max_no_improvement=max_no_improvement,
            reassignment_ratio=reassignment_ratio,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.cluster._mean_shift import MeanShift as _MeanShift


@nice_repr
class MeanShift(_MeanShift, SklearnEstimator):
    def __init__(self, bin_seeding: Boolean(), cluster_all: Boolean()):
        SklearnEstimator.__init__(self)
        _MeanShift.__init__(self, bin_seeding=bin_seeding, cluster_all=cluster_all)

    def run(
        self, input: Tuple(MatrixContinuousDense(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.decomposition._factor_analysis import FactorAnalysis as _FactorAnalysis


@nice_repr
class FactorAnalysis(_FactorAnalysis, SklearnTransformer):
    def __init__(
        self,
        tol: Continuous(min=-0.093, max=0.094),
        svd_method: Categorical("lapack", "randomized"),
        iterated_power: Discrete(min=1, max=5),
    ):
        SklearnTransformer.__init__(self)
        _FactorAnalysis.__init__(
            self, tol=tol, svd_method=svd_method, iterated_power=iterated_power
        )

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.decomposition._fastica import FastICA as _FastICA


@nice_repr
class FastICA(_FastICA, SklearnTransformer):
    def __init__(
        self,
        algorithm: Categorical("deflation", "parallel"),
        whiten: Boolean(),
        fun: Categorical("cube", "exp", "logcosh"),
    ):
        SklearnTransformer.__init__(self)
        _FastICA.__init__(self, algorithm=algorithm, whiten=whiten, fun=fun)

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.decomposition._incremental_pca import IncrementalPCA as _IncrementalPCA


@nice_repr
class IncrementalPCA(_IncrementalPCA, SklearnTransformer):
    def __init__(self, whiten: Boolean()):
        SklearnTransformer.__init__(self)
        _IncrementalPCA.__init__(self, whiten=whiten)

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.decomposition._kernel_pca import KernelPCA as _KernelPCA


@nice_repr
class KernelPCA(_KernelPCA, SklearnTransformer):
    def __init__(
        self,
        degree: Discrete(min=1, max=5),
        alpha: Continuous(min=-9.995, max=9.991),
        fit_inverse_transform: Boolean(),
        eigen_solver: Categorical("arpack", "auto", "dense"),
        tol: Discrete(min=-99, max=99),
        remove_zero_eig: Boolean(),
    ):
        SklearnTransformer.__init__(self)
        _KernelPCA.__init__(
            self,
            degree=degree,
            alpha=alpha,
            fit_inverse_transform=fit_inverse_transform,
            eigen_solver=eigen_solver,
            tol=tol,
            remove_zero_eig=remove_zero_eig,
        )

    def run(self, input: MatrixContinuous()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.decomposition._lda import (
    LatentDirichletAllocation as _LatentDirichletAllocation,
)


@nice_repr
class LatentDirichletAllocation(_LatentDirichletAllocation, SklearnTransformer):
    def __init__(self,):
        SklearnTransformer.__init__(self)
        _LatentDirichletAllocation.__init__(self,)

    def run(self, input: MatrixContinuous()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.decomposition._nmf import NMF as _NMF


@nice_repr
class NMF(_NMF, SklearnTransformer):
    def __init__(
        self,
        alpha: Continuous(min=0.0, max=0.0),
        l1_ratio: Continuous(min=0.0, max=0.0),
        shuffle: Boolean(),
    ):
        SklearnTransformer.__init__(self)
        _NMF.__init__(self, alpha=alpha, l1_ratio=l1_ratio, shuffle=shuffle)

    def run(self, input: MatrixContinuous()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.decomposition._pca import PCA as _PCA


@nice_repr
class PCA(_PCA, SklearnTransformer):
    def __init__(
        self,
        whiten: Boolean(),
        svd_solver: Categorical("arpack", "auto", "full", "randomized"),
        tol: Continuous(min=-0.992, max=0.992),
        iterated_power: Categorical("auto", "randomized"),
    ):
        SklearnTransformer.__init__(self)
        _PCA.__init__(
            self,
            whiten=whiten,
            svd_solver=svd_solver,
            tol=tol,
            iterated_power=iterated_power,
        )

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.decomposition._sparse_pca import MiniBatchSparsePCA as _MiniBatchSparsePCA


@nice_repr
class MiniBatchSparsePCA(_MiniBatchSparsePCA, SklearnTransformer):
    def __init__(
        self,
        ridge_alpha: Continuous(min=-0.093, max=0.094),
        n_iter: Discrete(min=1, max=199),
        batch_size: Discrete(min=1, max=5),
        shuffle: Boolean(),
        method: Categorical("cd", "lars"),
    ):
        SklearnTransformer.__init__(self)
        _MiniBatchSparsePCA.__init__(
            self,
            ridge_alpha=ridge_alpha,
            n_iter=n_iter,
            batch_size=batch_size,
            shuffle=shuffle,
            method=method,
        )

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.decomposition._sparse_pca import SparsePCA as _SparsePCA


@nice_repr
class SparsePCA(_SparsePCA, SklearnTransformer):
    def __init__(
        self,
        ridge_alpha: Continuous(min=-0.093, max=0.094),
        method: Categorical("cd", "lars"),
    ):
        SklearnTransformer.__init__(self)
        _SparsePCA.__init__(self, ridge_alpha=ridge_alpha, method=method)

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.decomposition._truncated_svd import TruncatedSVD as _TruncatedSVD


@nice_repr
class TruncatedSVD(_TruncatedSVD, SklearnTransformer):
    def __init__(
        self,
        n_components: Discrete(min=1, max=3),
        n_iter: Discrete(min=1, max=9),
        tol: Continuous(min=-0.992, max=0.992),
    ):
        SklearnTransformer.__init__(self)
        _TruncatedSVD.__init__(self, n_components=n_components, n_iter=n_iter, tol=tol)

    def run(self, input: MatrixContinuous()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.feature_extraction.text import CountVectorizer as _CountVectorizer


@nice_repr
class CountVectorizer(_CountVectorizer, SklearnTransformer):
    def __init__(self, lowercase: Boolean(), binary: Boolean()):
        SklearnTransformer.__init__(self)
        _CountVectorizer.__init__(self, lowercase=lowercase, binary=binary)

    def run(self, input: List(Sentence())) -> MatrixContinuousSparse():
        return SklearnTransformer.run(self, input)


from sklearn.feature_extraction.text import HashingVectorizer as _HashingVectorizer


@nice_repr
class HashingVectorizer(_HashingVectorizer, SklearnTransformer):
    def __init__(
        self,
        lowercase: Boolean(),
        n_features: Discrete(min=1, max=2097151),
        binary: Boolean(),
        norm: Categorical("l1"),
        alternate_sign: Boolean(),
    ):
        SklearnTransformer.__init__(self)
        _HashingVectorizer.__init__(
            self,
            lowercase=lowercase,
            n_features=n_features,
            binary=binary,
            norm=norm,
            alternate_sign=alternate_sign,
        )

    def run(self, input: List(Sentence())) -> MatrixContinuousSparse():
        return SklearnTransformer.run(self, input)


from sklearn.feature_extraction.text import TfidfTransformer as _TfidfTransformer


@nice_repr
class TfidfTransformer(_TfidfTransformer, SklearnTransformer):
    def __init__(
        self,
        norm: Categorical("l1", "l2"),
        use_idf: Boolean(),
        smooth_idf: Boolean(),
        sublinear_tf: Boolean(),
    ):
        SklearnTransformer.__init__(self)
        _TfidfTransformer.__init__(
            self,
            norm=norm,
            use_idf=use_idf,
            smooth_idf=smooth_idf,
            sublinear_tf=sublinear_tf,
        )

    def run(self, input: MatrixContinuous()) -> MatrixContinuousSparse():
        return SklearnTransformer.run(self, input)


from sklearn.feature_extraction.text import TfidfVectorizer as _TfidfVectorizer


@nice_repr
class TfidfVectorizer(_TfidfVectorizer, SklearnTransformer):
    def __init__(
        self,
        lowercase: Boolean(),
        binary: Boolean(),
        use_idf: Boolean(),
        smooth_idf: Boolean(),
        sublinear_tf: Boolean(),
    ):
        SklearnTransformer.__init__(self)
        _TfidfVectorizer.__init__(
            self,
            lowercase=lowercase,
            binary=binary,
            use_idf=use_idf,
            smooth_idf=smooth_idf,
            sublinear_tf=sublinear_tf,
        )

    def run(self, input: List(Sentence())) -> MatrixContinuousSparse():
        return SklearnTransformer.run(self, input)


from sklearn.impute._knn import KNNImputer as _KNNImputer


@nice_repr
class KNNImputer(_KNNImputer, SklearnTransformer):
    def __init__(
        self,
        n_neighbors: Discrete(min=1, max=9),
        weights: Categorical("distance", "uniform"),
        metric: Categorical("nan_euclidean"),
        add_indicator: Boolean(),
    ):
        SklearnTransformer.__init__(self)
        _KNNImputer.__init__(
            self,
            n_neighbors=n_neighbors,
            weights=weights,
            metric=metric,
            add_indicator=add_indicator,
        )

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.linear_model._base import LinearRegression as _LinearRegression


@nice_repr
class LinearRegression(_LinearRegression, SklearnEstimator):
    def __init__(self, fit_intercept: Boolean(), normalize: Boolean()):
        SklearnEstimator.__init__(self)
        _LinearRegression.__init__(
            self, fit_intercept=fit_intercept, normalize=normalize
        )

    def run(
        self, input: Tuple(MatrixContinuous(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._bayes import ARDRegression as _ARDRegression


@nice_repr
class ARDRegression(_ARDRegression, SklearnEstimator):
    def __init__(
        self,
        n_iter: Discrete(min=1, max=599),
        tol: Continuous(min=-0.005, max=0.001),
        compute_score: Boolean(),
        threshold_lambda: Continuous(min=-99999.993, max=99999.995),
        fit_intercept: Boolean(),
        normalize: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _ARDRegression.__init__(
            self,
            n_iter=n_iter,
            tol=tol,
            compute_score=compute_score,
            threshold_lambda=threshold_lambda,
            fit_intercept=fit_intercept,
            normalize=normalize,
        )

    def run(
        self, input: Tuple(MatrixContinuousDense(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._bayes import BayesianRidge as _BayesianRidge


@nice_repr
class BayesianRidge(_BayesianRidge, SklearnEstimator):
    def __init__(
        self,
        n_iter: Discrete(min=1, max=599),
        tol: Continuous(min=-0.005, max=0.001),
        compute_score: Boolean(),
        fit_intercept: Boolean(),
        normalize: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _BayesianRidge.__init__(
            self,
            n_iter=n_iter,
            tol=tol,
            compute_score=compute_score,
            fit_intercept=fit_intercept,
            normalize=normalize,
        )

    def run(
        self, input: Tuple(MatrixContinuousDense(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._coordinate_descent import ElasticNet as _ElasticNet


@nice_repr
class ElasticNet(_ElasticNet, SklearnEstimator):
    def __init__(
        self,
        alpha: Continuous(min=-9.995, max=9.991),
        l1_ratio: Continuous(min=-4.995, max=4.991),
        fit_intercept: Boolean(),
        normalize: Boolean(),
        precompute: Boolean(),
        positive: Boolean(),
        selection: Categorical("cyclic", "random"),
    ):
        SklearnEstimator.__init__(self)
        _ElasticNet.__init__(
            self,
            alpha=alpha,
            l1_ratio=l1_ratio,
            fit_intercept=fit_intercept,
            normalize=normalize,
            precompute=precompute,
            positive=positive,
            selection=selection,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._coordinate_descent import Lasso as _Lasso


@nice_repr
class Lasso(_Lasso, SklearnEstimator):
    def __init__(
        self,
        alpha: Continuous(min=-9.995, max=9.991),
        fit_intercept: Boolean(),
        normalize: Boolean(),
        precompute: Boolean(),
        positive: Boolean(),
        selection: Categorical("cyclic", "random"),
    ):
        SklearnEstimator.__init__(self)
        _Lasso.__init__(
            self,
            alpha=alpha,
            fit_intercept=fit_intercept,
            normalize=normalize,
            precompute=precompute,
            positive=positive,
            selection=selection,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._huber import HuberRegressor as _HuberRegressor


@nice_repr
class HuberRegressor(_HuberRegressor, SklearnEstimator):
    def __init__(
        self, epsilon: Continuous(min=1.002, max=13.494), fit_intercept: Boolean()
    ):
        SklearnEstimator.__init__(self)
        _HuberRegressor.__init__(self, epsilon=epsilon, fit_intercept=fit_intercept)

    def run(
        self, input: Tuple(MatrixContinuous(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._least_angle import Lars as _Lars


@nice_repr
class Lars(_Lars, SklearnEstimator):
    def __init__(
        self,
        fit_intercept: Boolean(),
        normalize: Boolean(),
        precompute: Categorical("auto"),
        n_nonzero_coefs: Discrete(min=1, max=999),
        fit_path: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _Lars.__init__(
            self,
            fit_intercept=fit_intercept,
            normalize=normalize,
            precompute=precompute,
            n_nonzero_coefs=n_nonzero_coefs,
            fit_path=fit_path,
        )

    def run(
        self, input: Tuple(MatrixContinuousDense(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._least_angle import LassoLars as _LassoLars


@nice_repr
class LassoLars(_LassoLars, SklearnEstimator):
    def __init__(
        self,
        fit_intercept: Boolean(),
        normalize: Boolean(),
        precompute: Categorical("auto"),
        fit_path: Boolean(),
        positive: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _LassoLars.__init__(
            self,
            fit_intercept=fit_intercept,
            normalize=normalize,
            precompute=precompute,
            fit_path=fit_path,
            positive=positive,
        )

    def run(
        self, input: Tuple(MatrixContinuousDense(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._least_angle import LassoLarsIC as _LassoLarsIC


@nice_repr
class LassoLarsIC(_LassoLarsIC, SklearnEstimator):
    def __init__(
        self,
        criterion: Categorical("aic", "bic"),
        fit_intercept: Boolean(),
        normalize: Boolean(),
        precompute: Categorical("auto"),
        positive: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _LassoLarsIC.__init__(
            self,
            criterion=criterion,
            fit_intercept=fit_intercept,
            normalize=normalize,
            precompute=precompute,
            positive=positive,
        )

    def run(
        self, input: Tuple(MatrixContinuousDense(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._logistic import LogisticRegression as _LogisticRegression


@nice_repr
class LogisticRegression(_LogisticRegression, SklearnEstimator):
    def __init__(
        self,
        penalty: Categorical("l2", "none"),
        dual: Boolean(),
        C: Continuous(min=0.005, max=9.991),
        fit_intercept: Boolean(),
        solver: Categorical("lbfgs", "liblinear", "sag", "saga"),
        multi_class: Categorical("auto", "multinomial", "ovr"),
    ):
        SklearnEstimator.__init__(self)
        _LogisticRegression.__init__(
            self,
            penalty=penalty,
            dual=dual,
            C=C,
            fit_intercept=fit_intercept,
            solver=solver,
            multi_class=multi_class,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._omp import (
    OrthogonalMatchingPursuit as _OrthogonalMatchingPursuit,
)


@nice_repr
class OrthogonalMatchingPursuit(_OrthogonalMatchingPursuit, SklearnEstimator):
    def __init__(
        self,
        fit_intercept: Boolean(),
        normalize: Boolean(),
        precompute: Categorical("auto"),
    ):
        SklearnEstimator.__init__(self)
        _OrthogonalMatchingPursuit.__init__(
            self,
            fit_intercept=fit_intercept,
            normalize=normalize,
            precompute=precompute,
        )

    def run(
        self, input: Tuple(MatrixContinuousDense(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._passive_aggressive import (
    PassiveAggressiveClassifier as _PassiveAggressiveClassifier,
)


@nice_repr
class PassiveAggressiveClassifier(_PassiveAggressiveClassifier, SklearnEstimator):
    def __init__(
        self,
        C: Continuous(min=-9.995, max=9.991),
        fit_intercept: Boolean(),
        tol: Continuous(min=-0.005, max=0.001),
        early_stopping: Boolean(),
        validation_fraction: Continuous(min=0.006, max=0.993),
        n_iter_no_change: Discrete(min=1, max=9),
        shuffle: Boolean(),
        average: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _PassiveAggressiveClassifier.__init__(
            self,
            C=C,
            fit_intercept=fit_intercept,
            tol=tol,
            early_stopping=early_stopping,
            validation_fraction=validation_fraction,
            n_iter_no_change=n_iter_no_change,
            shuffle=shuffle,
            average=average,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._passive_aggressive import (
    PassiveAggressiveRegressor as _PassiveAggressiveRegressor,
)


@nice_repr
class PassiveAggressiveRegressor(_PassiveAggressiveRegressor, SklearnEstimator):
    def __init__(
        self,
        C: Continuous(min=-9.995, max=9.991),
        fit_intercept: Boolean(),
        tol: Continuous(min=-0.005, max=0.001),
        early_stopping: Boolean(),
        validation_fraction: Continuous(min=0.006, max=0.993),
        n_iter_no_change: Discrete(min=1, max=9),
        shuffle: Boolean(),
        epsilon: Continuous(min=-0.992, max=0.993),
        average: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _PassiveAggressiveRegressor.__init__(
            self,
            C=C,
            fit_intercept=fit_intercept,
            tol=tol,
            early_stopping=early_stopping,
            validation_fraction=validation_fraction,
            n_iter_no_change=n_iter_no_change,
            shuffle=shuffle,
            epsilon=epsilon,
            average=average,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._perceptron import Perceptron as _Perceptron


@nice_repr
class Perceptron(_Perceptron, SklearnEstimator):
    def __init__(
        self,
        fit_intercept: Boolean(),
        tol: Continuous(min=-0.005, max=0.001),
        shuffle: Boolean(),
        eta0: Continuous(min=0.005, max=9.991),
        early_stopping: Boolean(),
        validation_fraction: Continuous(min=0.006, max=0.993),
        n_iter_no_change: Discrete(min=1, max=9),
    ):
        SklearnEstimator.__init__(self)
        _Perceptron.__init__(
            self,
            fit_intercept=fit_intercept,
            tol=tol,
            shuffle=shuffle,
            eta0=eta0,
            early_stopping=early_stopping,
            validation_fraction=validation_fraction,
            n_iter_no_change=n_iter_no_change,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._ridge import Ridge as _Ridge


@nice_repr
class Ridge(_Ridge, SklearnEstimator):
    def __init__(
        self,
        alpha: Continuous(min=-9.995, max=9.991),
        fit_intercept: Boolean(),
        normalize: Boolean(),
        tol: Continuous(min=-0.005, max=0.001),
        solver: Categorical(
            "auto", "cholesky", "lsqr", "sag", "saga", "sparse_cg", "svd"
        ),
    ):
        SklearnEstimator.__init__(self)
        _Ridge.__init__(
            self,
            alpha=alpha,
            fit_intercept=fit_intercept,
            normalize=normalize,
            tol=tol,
            solver=solver,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._ridge import RidgeClassifier as _RidgeClassifier


@nice_repr
class RidgeClassifier(_RidgeClassifier, SklearnEstimator):
    def __init__(
        self,
        alpha: Continuous(min=-9.995, max=9.991),
        fit_intercept: Boolean(),
        normalize: Boolean(),
        tol: Continuous(min=-0.005, max=0.001),
        solver: Categorical(
            "auto", "cholesky", "lsqr", "sag", "saga", "sparse_cg", "svd"
        ),
    ):
        SklearnEstimator.__init__(self)
        _RidgeClassifier.__init__(
            self,
            alpha=alpha,
            fit_intercept=fit_intercept,
            normalize=normalize,
            tol=tol,
            solver=solver,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._stochastic_gradient import SGDClassifier as _SGDClassifier


@nice_repr
class SGDClassifier(_SGDClassifier, SklearnEstimator):
    def __init__(
        self,
        loss: Categorical(
            "epsilon_insensitive",
            "hinge",
            "huber",
            "log",
            "modified_huber",
            "perceptron",
            "squared_epsilon_insensitive",
            "squared_hinge",
            "squared_loss",
        ),
        penalty: Categorical("elasticnet", "l1", "l2"),
        l1_ratio: Continuous(min=0.001, max=0.999),
        fit_intercept: Boolean(),
        tol: Continuous(min=-0.005, max=0.001),
        shuffle: Boolean(),
        epsilon: Continuous(min=-0.992, max=0.993),
        learning_rate: Categorical("optimal"),
        eta0: Continuous(min=-0.992, max=0.992),
        power_t: Continuous(min=-4.995, max=4.991),
        early_stopping: Boolean(),
        validation_fraction: Continuous(min=0.006, max=0.993),
        n_iter_no_change: Discrete(min=1, max=9),
        average: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _SGDClassifier.__init__(
            self,
            loss=loss,
            penalty=penalty,
            l1_ratio=l1_ratio,
            fit_intercept=fit_intercept,
            tol=tol,
            shuffle=shuffle,
            epsilon=epsilon,
            learning_rate=learning_rate,
            eta0=eta0,
            power_t=power_t,
            early_stopping=early_stopping,
            validation_fraction=validation_fraction,
            n_iter_no_change=n_iter_no_change,
            average=average,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._stochastic_gradient import SGDRegressor as _SGDRegressor


@nice_repr
class SGDRegressor(_SGDRegressor, SklearnEstimator):
    def __init__(
        self,
        loss: Categorical(
            "epsilon_insensitive",
            "huber",
            "squared_epsilon_insensitive",
            "squared_loss",
        ),
        penalty: Categorical("elasticnet", "l1", "l2"),
        l1_ratio: Continuous(min=0.001, max=0.999),
        fit_intercept: Boolean(),
        tol: Continuous(min=-0.005, max=0.001),
        shuffle: Boolean(),
        epsilon: Continuous(min=-0.992, max=0.993),
        learning_rate: Categorical("adaptive", "constant", "invscaling", "optimal"),
        eta0: Continuous(min=0.003, max=0.094),
        power_t: Continuous(min=-2.494, max=2.491),
        early_stopping: Boolean(),
        validation_fraction: Continuous(min=0.006, max=0.993),
        n_iter_no_change: Discrete(min=1, max=9),
        average: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _SGDRegressor.__init__(
            self,
            loss=loss,
            penalty=penalty,
            l1_ratio=l1_ratio,
            fit_intercept=fit_intercept,
            tol=tol,
            shuffle=shuffle,
            epsilon=epsilon,
            learning_rate=learning_rate,
            eta0=eta0,
            power_t=power_t,
            early_stopping=early_stopping,
            validation_fraction=validation_fraction,
            n_iter_no_change=n_iter_no_change,
            average=average,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.linear_model._theil_sen import TheilSenRegressor as _TheilSenRegressor


@nice_repr
class TheilSenRegressor(_TheilSenRegressor, SklearnEstimator):
    def __init__(
        self, fit_intercept: Boolean(), tol: Continuous(min=-0.005, max=0.001)
    ):
        SklearnEstimator.__init__(self)
        _TheilSenRegressor.__init__(self, fit_intercept=fit_intercept, tol=tol)

    def run(
        self, input: Tuple(MatrixContinuousDense(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.manifold._isomap import Isomap as _Isomap


@nice_repr
class Isomap(_Isomap, SklearnTransformer):
    def __init__(
        self,
        n_neighbors: Discrete(min=1, max=9),
        n_components: Discrete(min=1, max=3),
        eigen_solver: Categorical("arpack", "auto", "dense"),
        tol: Discrete(min=-99, max=99),
        path_method: Categorical("auto"),
        neighbors_algorithm: Categorical("auto", "ball_tree", "brute", "kd_tree"),
        p: Discrete(min=1, max=3),
    ):
        SklearnTransformer.__init__(self)
        _Isomap.__init__(
            self,
            n_neighbors=n_neighbors,
            n_components=n_components,
            eigen_solver=eigen_solver,
            tol=tol,
            path_method=path_method,
            neighbors_algorithm=neighbors_algorithm,
            p=p,
        )

    def run(self, input: MatrixContinuous()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.manifold._locally_linear import (
    LocallyLinearEmbedding as _LocallyLinearEmbedding,
)


@nice_repr
class LocallyLinearEmbedding(_LocallyLinearEmbedding, SklearnTransformer):
    def __init__(
        self,
        n_neighbors: Discrete(min=1, max=9),
        n_components: Discrete(min=1, max=3),
        reg: Continuous(min=-0.005, max=0.001),
        eigen_solver: Categorical("arpack", "auto", "dense"),
        method: Categorical("ltsa", "modified", "standard"),
        neighbors_algorithm: Categorical("auto", "ball_tree", "brute", "kd_tree"),
    ):
        SklearnTransformer.__init__(self)
        _LocallyLinearEmbedding.__init__(
            self,
            n_neighbors=n_neighbors,
            n_components=n_components,
            reg=reg,
            eigen_solver=eigen_solver,
            method=method,
            neighbors_algorithm=neighbors_algorithm,
        )

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.naive_bayes import BernoulliNB as _BernoulliNB


@nice_repr
class BernoulliNB(_BernoulliNB, SklearnEstimator):
    def __init__(
        self,
        alpha: Continuous(min=0.0, max=9.991),
        binarize: Continuous(min=-0.992, max=0.992),
        fit_prior: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _BernoulliNB.__init__(self, alpha=alpha, binarize=binarize, fit_prior=fit_prior)

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.naive_bayes import CategoricalNB as _CategoricalNB


@nice_repr
class CategoricalNB(_CategoricalNB, SklearnEstimator):
    def __init__(self, fit_prior: Boolean()):
        SklearnEstimator.__init__(self)
        _CategoricalNB.__init__(self, fit_prior=fit_prior)

    def run(
        self, input: Tuple(MatrixContinuousDense(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.naive_bayes import ComplementNB as _ComplementNB


@nice_repr
class ComplementNB(_ComplementNB, SklearnEstimator):
    def __init__(self, fit_prior: Boolean(), norm: Boolean()):
        SklearnEstimator.__init__(self)
        _ComplementNB.__init__(self, fit_prior=fit_prior, norm=norm)

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.naive_bayes import GaussianNB as _GaussianNB


@nice_repr
class GaussianNB(_GaussianNB, SklearnEstimator):
    def __init__(self,):
        SklearnEstimator.__init__(self)
        _GaussianNB.__init__(self,)

    def run(
        self, input: Tuple(MatrixContinuousDense(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.naive_bayes import MultinomialNB as _MultinomialNB


@nice_repr
class MultinomialNB(_MultinomialNB, SklearnEstimator):
    def __init__(self, fit_prior: Boolean()):
        SklearnEstimator.__init__(self)
        _MultinomialNB.__init__(self, fit_prior=fit_prior)

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.neighbors._classification import (
    KNeighborsClassifier as _KNeighborsClassifier,
)


@nice_repr
class KNeighborsClassifier(_KNeighborsClassifier, SklearnEstimator):
    def __init__(
        self,
        n_neighbors: Discrete(min=1, max=9),
        weights: Categorical("distance", "uniform"),
        algorithm: Categorical("auto", "ball_tree", "brute", "kd_tree"),
        leaf_size: Discrete(min=1, max=59),
        p: Discrete(min=1, max=3),
        metric: Categorical("minkowski"),
    ):
        SklearnEstimator.__init__(self)
        _KNeighborsClassifier.__init__(
            self,
            n_neighbors=n_neighbors,
            weights=weights,
            algorithm=algorithm,
            leaf_size=leaf_size,
            p=p,
            metric=metric,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.neighbors._graph import KNeighborsTransformer as _KNeighborsTransformer


@nice_repr
class KNeighborsTransformer(_KNeighborsTransformer, SklearnTransformer):
    def __init__(
        self,
        mode: Categorical("connectivity", "distance"),
        n_neighbors: Discrete(min=1, max=9),
        algorithm: Categorical("auto", "ball_tree", "brute", "kd_tree"),
        leaf_size: Discrete(min=1, max=59),
        p: Discrete(min=1, max=3),
    ):
        SklearnTransformer.__init__(self)
        _KNeighborsTransformer.__init__(
            self,
            mode=mode,
            n_neighbors=n_neighbors,
            algorithm=algorithm,
            leaf_size=leaf_size,
            p=p,
        )

    def run(self, input: MatrixContinuous()) -> MatrixContinuousSparse():
        return SklearnTransformer.run(self, input)


from sklearn.neighbors._graph import (
    RadiusNeighborsTransformer as _RadiusNeighborsTransformer,
)


@nice_repr
class RadiusNeighborsTransformer(_RadiusNeighborsTransformer, SklearnTransformer):
    def __init__(
        self,
        mode: Categorical("connectivity", "distance"),
        radius: Continuous(min=-9.995, max=9.991),
        algorithm: Categorical("auto", "ball_tree", "brute", "kd_tree"),
        leaf_size: Discrete(min=1, max=59),
        p: Discrete(min=1, max=3),
    ):
        SklearnTransformer.__init__(self)
        _RadiusNeighborsTransformer.__init__(
            self,
            mode=mode,
            radius=radius,
            algorithm=algorithm,
            leaf_size=leaf_size,
            p=p,
        )

    def run(self, input: MatrixContinuous()) -> MatrixContinuousSparse():
        return SklearnTransformer.run(self, input)


from sklearn.neighbors._lof import LocalOutlierFactor as _LocalOutlierFactor


@nice_repr
class LocalOutlierFactor(_LocalOutlierFactor, SklearnEstimator):
    def __init__(
        self,
        n_neighbors: Discrete(min=1, max=39),
        algorithm: Categorical("auto", "ball_tree", "brute", "kd_tree"),
        leaf_size: Discrete(min=1, max=59),
        p: Discrete(min=1, max=3),
        contamination: Categorical("auto"),
        novelty: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _LocalOutlierFactor.__init__(
            self,
            n_neighbors=n_neighbors,
            algorithm=algorithm,
            leaf_size=leaf_size,
            p=p,
            contamination=contamination,
            novelty=novelty,
        )

    def run(self, input: MatrixContinuous()) -> DiscreteVector():
        return SklearnEstimator.run(self, input)


from sklearn.neighbors._nearest_centroid import NearestCentroid as _NearestCentroid


@nice_repr
class NearestCentroid(_NearestCentroid, SklearnEstimator):
    def __init__(self,):
        SklearnEstimator.__init__(self)
        _NearestCentroid.__init__(self,)

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.neighbors._regression import KNeighborsRegressor as _KNeighborsRegressor


@nice_repr
class KNeighborsRegressor(_KNeighborsRegressor, SklearnEstimator):
    def __init__(
        self,
        n_neighbors: Discrete(min=1, max=9),
        weights: Categorical("distance", "uniform"),
        algorithm: Categorical("auto", "ball_tree", "brute", "kd_tree"),
        leaf_size: Discrete(min=1, max=59),
        p: Discrete(min=1, max=3),
        metric: Categorical("minkowski"),
    ):
        SklearnEstimator.__init__(self)
        _KNeighborsRegressor.__init__(
            self,
            n_neighbors=n_neighbors,
            weights=weights,
            algorithm=algorithm,
            leaf_size=leaf_size,
            p=p,
            metric=metric,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.neighbors._regression import (
    RadiusNeighborsRegressor as _RadiusNeighborsRegressor,
)


@nice_repr
class RadiusNeighborsRegressor(_RadiusNeighborsRegressor, SklearnEstimator):
    def __init__(
        self,
        radius: Continuous(min=-9.995, max=9.991),
        weights: Categorical("distance", "uniform"),
        algorithm: Categorical("auto", "ball_tree", "brute", "kd_tree"),
        leaf_size: Discrete(min=1, max=59),
        p: Discrete(min=1, max=3),
    ):
        SklearnEstimator.__init__(self)
        _RadiusNeighborsRegressor.__init__(
            self,
            radius=radius,
            weights=weights,
            algorithm=algorithm,
            leaf_size=leaf_size,
            p=p,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.preprocessing._data import KernelCenterer as _KernelCenterer


@nice_repr
class KernelCenterer(_KernelCenterer, SklearnTransformer):
    def __init__(self,):
        SklearnTransformer.__init__(self)
        _KernelCenterer.__init__(self,)

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.preprocessing._data import MinMaxScaler as _MinMaxScaler


@nice_repr
class MinMaxScaler(_MinMaxScaler, SklearnTransformer):
    def __init__(self,):
        SklearnTransformer.__init__(self)
        _MinMaxScaler.__init__(self,)

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.preprocessing._data import PowerTransformer as _PowerTransformer


@nice_repr
class PowerTransformer(_PowerTransformer, SklearnTransformer):
    def __init__(self, standardize: Boolean()):
        SklearnTransformer.__init__(self)
        _PowerTransformer.__init__(self, standardize=standardize)

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.preprocessing._data import RobustScaler as _RobustScaler


@nice_repr
class RobustScaler(_RobustScaler, SklearnTransformer):
    def __init__(self, with_centering: Boolean(), with_scaling: Boolean()):
        SklearnTransformer.__init__(self)
        _RobustScaler.__init__(
            self, with_centering=with_centering, with_scaling=with_scaling
        )

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.preprocessing._data import StandardScaler as _StandardScaler


@nice_repr
class StandardScaler(_StandardScaler, SklearnTransformer):
    def __init__(self, with_mean: Boolean(), with_std: Boolean()):
        SklearnTransformer.__init__(self)
        _StandardScaler.__init__(self, with_mean=with_mean, with_std=with_std)

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.preprocessing._discretization import KBinsDiscretizer as _KBinsDiscretizer


@nice_repr
class KBinsDiscretizer(_KBinsDiscretizer, SklearnTransformer):
    def __init__(
        self,
        n_bins: Discrete(min=2, max=9),
        encode: Categorical("onehot", "ordinal"),
        strategy: Categorical("kmeans", "quantile", "uniform"),
    ):
        SklearnTransformer.__init__(self)
        _KBinsDiscretizer.__init__(
            self, n_bins=n_bins, encode=encode, strategy=strategy
        )

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousSparse():
        return SklearnTransformer.run(self, input)


from sklearn.preprocessing._encoders import OneHotEncoder as _OneHotEncoder


@nice_repr
class OneHotEncoder(_OneHotEncoder, SklearnTransformer):
    def __init__(
        self,
        categories: Categorical("auto"),
        sparse: Boolean(),
        handle_unknown: Categorical("error", "ignore"),
    ):
        SklearnTransformer.__init__(self)
        _OneHotEncoder.__init__(
            self, categories=categories, sparse=sparse, handle_unknown=handle_unknown
        )

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousSparse():
        return SklearnTransformer.run(self, input)


from sklearn.preprocessing._encoders import OrdinalEncoder as _OrdinalEncoder


@nice_repr
class OrdinalEncoder(_OrdinalEncoder, SklearnTransformer):
    def __init__(self, categories: Categorical("auto")):
        SklearnTransformer.__init__(self)
        _OrdinalEncoder.__init__(self, categories=categories)

    def run(self, input: MatrixContinuousDense()) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.preprocessing._label import LabelBinarizer as _LabelBinarizer


@nice_repr
class LabelBinarizer(_LabelBinarizer, SklearnTransformer):
    def __init__(self, sparse_output: Boolean()):
        SklearnTransformer.__init__(self)
        _LabelBinarizer.__init__(self, sparse_output=sparse_output)

    def run(self, input: List(Category())) -> MatrixContinuousDense():
        return SklearnTransformer.run(self, input)


from sklearn.svm._classes import LinearSVC as _LinearSVC


@nice_repr
class LinearSVC(_LinearSVC, SklearnEstimator):
    def __init__(
        self,
        penalty: Categorical("l2"),
        loss: Categorical("hinge", "squared_hinge"),
        dual: Boolean(),
        C: Continuous(min=0.005, max=9.991),
        multi_class: Categorical("crammer_singer", "ovr"),
        fit_intercept: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _LinearSVC.__init__(
            self,
            penalty=penalty,
            loss=loss,
            dual=dual,
            C=C,
            multi_class=multi_class,
            fit_intercept=fit_intercept,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.svm._classes import LinearSVR as _LinearSVR


@nice_repr
class LinearSVR(_LinearSVR, SklearnEstimator):
    def __init__(
        self,
        epsilon: Continuous(min=0.0, max=0.992),
        C: Continuous(min=0.005, max=9.991),
        loss: Categorical("epsilon_insensitive", "squared_epsilon_insensitive"),
        fit_intercept: Boolean(),
        intercept_scaling: Continuous(min=0.005, max=9.991),
        dual: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _LinearSVR.__init__(
            self,
            epsilon=epsilon,
            C=C,
            loss=loss,
            fit_intercept=fit_intercept,
            intercept_scaling=intercept_scaling,
            dual=dual,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.svm._classes import NuSVC as _NuSVC


@nice_repr
class NuSVC(_NuSVC, SklearnEstimator):
    def __init__(
        self,
        kernel: Categorical("linear", "poly", "rbf", "sigmoid"),
        degree: Discrete(min=1, max=5),
        gamma: Categorical("auto", "scale"),
        coef0: Continuous(min=-0.992, max=0.992),
        shrinking: Boolean(),
        probability: Boolean(),
        cache_size: Discrete(min=1, max=399),
        decision_function_shape: Categorical("ovo", "ovr"),
        break_ties: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _NuSVC.__init__(
            self,
            kernel=kernel,
            degree=degree,
            gamma=gamma,
            coef0=coef0,
            shrinking=shrinking,
            probability=probability,
            cache_size=cache_size,
            decision_function_shape=decision_function_shape,
            break_ties=break_ties,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.svm._classes import NuSVR as _NuSVR


@nice_repr
class NuSVR(_NuSVR, SklearnEstimator):
    def __init__(
        self,
        C: Continuous(min=0.005, max=9.991),
        kernel: Categorical("linear", "poly", "rbf", "sigmoid"),
        degree: Discrete(min=1, max=5),
        gamma: Categorical("auto", "scale"),
        coef0: Continuous(min=-0.992, max=0.992),
        shrinking: Boolean(),
        cache_size: Discrete(min=1, max=399),
    ):
        SklearnEstimator.__init__(self)
        _NuSVR.__init__(
            self,
            C=C,
            kernel=kernel,
            degree=degree,
            gamma=gamma,
            coef0=coef0,
            shrinking=shrinking,
            cache_size=cache_size,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.svm._classes import OneClassSVM as _OneClassSVM


@nice_repr
class OneClassSVM(_OneClassSVM, SklearnEstimator):
    def __init__(
        self,
        kernel: Categorical("linear", "poly", "rbf", "sigmoid"),
        degree: Discrete(min=1, max=5),
        gamma: Categorical("auto", "scale"),
        coef0: Continuous(min=-0.992, max=0.992),
        shrinking: Boolean(),
        cache_size: Discrete(min=1, max=399),
    ):
        SklearnEstimator.__init__(self)
        _OneClassSVM.__init__(
            self,
            kernel=kernel,
            degree=degree,
            gamma=gamma,
            coef0=coef0,
            shrinking=shrinking,
            cache_size=cache_size,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.svm._classes import SVC as _SVC


@nice_repr
class SVC(_SVC, SklearnEstimator):
    def __init__(
        self,
        C: Continuous(min=0.005, max=9.991),
        degree: Discrete(min=1, max=5),
        gamma: Categorical("auto", "scale"),
        coef0: Continuous(min=-0.992, max=0.992),
        shrinking: Boolean(),
        probability: Boolean(),
        cache_size: Discrete(min=1, max=399),
        decision_function_shape: Categorical("ovo", "ovr"),
        break_ties: Boolean(),
    ):
        SklearnEstimator.__init__(self)
        _SVC.__init__(
            self,
            C=C,
            degree=degree,
            gamma=gamma,
            coef0=coef0,
            shrinking=shrinking,
            probability=probability,
            cache_size=cache_size,
            decision_function_shape=decision_function_shape,
            break_ties=break_ties,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.svm._classes import SVR as _SVR


@nice_repr
class SVR(_SVR, SklearnEstimator):
    def __init__(
        self,
        kernel: Categorical("linear", "poly", "rbf", "sigmoid"),
        degree: Discrete(min=1, max=5),
        gamma: Categorical("auto", "scale"),
        coef0: Continuous(min=-0.992, max=0.992),
        C: Continuous(min=0.005, max=9.991),
        epsilon: Continuous(min=0.006, max=0.993),
        shrinking: Boolean(),
        cache_size: Discrete(min=1, max=399),
    ):
        SklearnEstimator.__init__(self)
        _SVR.__init__(
            self,
            kernel=kernel,
            degree=degree,
            gamma=gamma,
            coef0=coef0,
            C=C,
            epsilon=epsilon,
            shrinking=shrinking,
            cache_size=cache_size,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.tree._classes import DecisionTreeClassifier as _DecisionTreeClassifier


@nice_repr
class DecisionTreeClassifier(_DecisionTreeClassifier, SklearnEstimator):
    def __init__(
        self,
        min_samples_split: Discrete(min=2, max=3),
        min_weight_fraction_leaf: Continuous(min=0.0, max=0.5),
        min_impurity_decrease: Continuous(min=0.0, max=0.992),
        ccp_alpha: Continuous(min=0.0, max=0.992),
    ):
        SklearnEstimator.__init__(self)
        _DecisionTreeClassifier.__init__(
            self,
            min_samples_split=min_samples_split,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            min_impurity_decrease=min_impurity_decrease,
            ccp_alpha=ccp_alpha,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.tree._classes import DecisionTreeRegressor as _DecisionTreeRegressor


@nice_repr
class DecisionTreeRegressor(_DecisionTreeRegressor, SklearnEstimator):
    def __init__(
        self,
        min_samples_split: Discrete(min=2, max=3),
        min_weight_fraction_leaf: Continuous(min=0.0, max=0.5),
        min_impurity_decrease: Continuous(min=0.0, max=0.992),
        ccp_alpha: Continuous(min=0.0, max=0.992),
    ):
        SklearnEstimator.__init__(self)
        _DecisionTreeRegressor.__init__(
            self,
            min_samples_split=min_samples_split,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            min_impurity_decrease=min_impurity_decrease,
            ccp_alpha=ccp_alpha,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


from sklearn.tree._classes import ExtraTreeClassifier as _ExtraTreeClassifier


@nice_repr
class ExtraTreeClassifier(_ExtraTreeClassifier, SklearnEstimator):
    def __init__(
        self,
        min_samples_split: Discrete(min=2, max=3),
        min_weight_fraction_leaf: Continuous(min=0.0, max=0.5),
        min_impurity_decrease: Continuous(min=0.0, max=0.992),
        ccp_alpha: Continuous(min=0.0, max=0.992),
    ):
        SklearnEstimator.__init__(self)
        _ExtraTreeClassifier.__init__(
            self,
            min_samples_split=min_samples_split,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            min_impurity_decrease=min_impurity_decrease,
            ccp_alpha=ccp_alpha,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), CategoricalVector())
    ) -> CategoricalVector():
        return SklearnEstimator.run(self, input)


from sklearn.tree._classes import ExtraTreeRegressor as _ExtraTreeRegressor


@nice_repr
class ExtraTreeRegressor(_ExtraTreeRegressor, SklearnEstimator):
    def __init__(
        self,
        min_samples_split: Discrete(min=2, max=3),
        min_weight_fraction_leaf: Continuous(min=0.0, max=0.5),
        min_impurity_decrease: Continuous(min=0.0, max=0.992),
        ccp_alpha: Continuous(min=0.0, max=0.992),
    ):
        SklearnEstimator.__init__(self)
        _ExtraTreeRegressor.__init__(
            self,
            min_samples_split=min_samples_split,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            min_impurity_decrease=min_impurity_decrease,
            ccp_alpha=ccp_alpha,
        )

    def run(
        self, input: Tuple(MatrixContinuous(), ContinuousVector())
    ) -> ContinuousVector():
        return SklearnEstimator.run(self, input)


__all__ = [
    "AffinityPropagation",
    "FeatureAgglomeration",
    "Birch",
    "KMeans",
    "MiniBatchKMeans",
    "MeanShift",
    "FactorAnalysis",
    "FastICA",
    "IncrementalPCA",
    "KernelPCA",
    "LatentDirichletAllocation",
    "NMF",
    "PCA",
    "MiniBatchSparsePCA",
    "SparsePCA",
    "TruncatedSVD",
    "CountVectorizer",
    "HashingVectorizer",
    "TfidfTransformer",
    "TfidfVectorizer",
    "KNNImputer",
    "LinearRegression",
    "ARDRegression",
    "BayesianRidge",
    "ElasticNet",
    "Lasso",
    "HuberRegressor",
    "Lars",
    "LassoLars",
    "LassoLarsIC",
    "LogisticRegression",
    "OrthogonalMatchingPursuit",
    "PassiveAggressiveClassifier",
    "PassiveAggressiveRegressor",
    "Perceptron",
    "Ridge",
    "RidgeClassifier",
    "SGDClassifier",
    "SGDRegressor",
    "TheilSenRegressor",
    "Isomap",
    "LocallyLinearEmbedding",
    "BernoulliNB",
    "CategoricalNB",
    "ComplementNB",
    "GaussianNB",
    "MultinomialNB",
    "KNeighborsClassifier",
    "KNeighborsTransformer",
    "RadiusNeighborsTransformer",
    "LocalOutlierFactor",
    "NearestCentroid",
    "KNeighborsRegressor",
    "RadiusNeighborsRegressor",
    "KernelCenterer",
    "MinMaxScaler",
    "PowerTransformer",
    "RobustScaler",
    "StandardScaler",
    "KBinsDiscretizer",
    "OneHotEncoder",
    "OrdinalEncoder",
    "LabelBinarizer",
    "LinearSVC",
    "LinearSVR",
    "NuSVC",
    "NuSVR",
    "OneClassSVM",
    "SVC",
    "SVR",
    "DecisionTreeClassifier",
    "DecisionTreeRegressor",
    "ExtraTreeClassifier",
    "ExtraTreeRegressor",
]



import enum
import inspect
import collections


MAX_REPR_DEPTH = 10

_repr_depth = [0]


def nice_repr(cls):
    """
    A decorator that adds a nice `repr(.)` to any decorated class.

    Decorate a class with `@nice_repr` to automatically generate a `__repr__()`
    method that prints the class name along with any parameters defined in the
    constructor which can be found in `dir(self)`.

    ##### Examples

    All of the parameters that you want to be printed in `repr(.)` should
    be either stored in the instance or accesible by name (e.g., as a property).

    ```python
    >>> @nice_repr
    ... class MyType:
    ...     def __init__(self, a, b, c):
    ...         self.a = a
    ...         self._b = b
    ...         self._c = c
    ...
    ...     @property
    ...     def b(self):
    ...         return self._b
    ...
    >>> x = MyType(42, b='hello', c='world')
    >>> x
    MyType(a=42, b="hello")

    ```

    It works nicely with nested objects, if all of them are `@nice_repr` decorated.

    ```python
    >>> @nice_repr
    ... class A:
    ...     def __init__(self, inner):
    ...         self.inner = inner
    >>> @nice_repr
    ... class B:
    ...     def __init__(self, value):
    ...         self.value = value
    >>> A([B(i) for i in range(10)])
    A(
        inner=[
            B(value=0),
            B(value=1),
            B(value=2),
            B(value=3),
            B(value=4),
            B(value=5),
            B(value=6),
            B(value=7),
            B(value=8),
            B(value=9),
        ]
    )

    ```

    It works with cyclic object graphs as well:

    ```python
    >>> @nice_repr
    ... class A:
    ...     def __init__(self, a:A=None):
    ...         self.a = self
    >>> A()
    A(a=A(a=A(a=A(a=A(a=A(a=A(a=A(a=A(a=A(a=A(a=A(...))))))))))))

    ```

    !!! note
        Change `autogoal.utils.MAX_REPR_DEPTH` to increase the depth level of recursive `repr`.

    """
    def repr_method(self):
        init_signature = inspect.signature(self.__init__)
        exclude_param_names = set(['self'])

        if _repr_depth[0] > MAX_REPR_DEPTH:
            return f"{self.__class__.__name__}(...)"

        _repr_depth[0] += 1

        parameter_names = [name for name in init_signature.parameters if name not in exclude_param_names]
        parameter_values = [getattr(self, param, None) for param in parameter_names]

        if hasattr(self, '__nice_repr_hook__'):
            self.__nice_repr_hook__(parameter_names, parameter_values)

        args = ", ".join(f"{name}={repr(value)}" for name, value in zip(parameter_names, parameter_values) if value is not None)
        fr = f"{self.__class__.__name__}({args})"

        _repr_depth[0] -= 1

        try:
            import black
            return black.format_str(fr, mode=black.FileMode()).strip()
        except:
            return fr

    cls.__repr__ = repr_method
    return cls


Kb = 1024
Mb = 1024 * Kb
Gb = 1024 * Mb

Sec = 1
Min = 60 * Sec
Hour = 60 * Min


def flatten(y):
    """
    Recursively flattens a list.

    ##### Examples

    ```python
    >>> flatten([[1],[2,[3]],4])
    [1, 2, 3, 4]

    ```
    """
    if isinstance(y, list):
        return [z for x in y for z in flatten(x)]
    else:
        return [y]


def compute_class_weights(y):
    """
    Computes relative class weights for imbalanced datasets. Works with nested input.

    ##### Examples

    ```python
    >>> compute_class_weights([['A', 'B', 'A'], ['C'], ['C', 'C']])
    {'A': 1.5, 'B': 3.0, 'C': 1.0}

    ```
    """
    y = flatten(y)
    class_counts = collections.Counter(y)
    _, max_class = class_counts.most_common(1)[0]

    return {k: max_class / v for k, v in class_counts.items()}


def factory(func_or_type, *args, **kwargs):
    def call():
        return func_or_type(*args, **kwargs)

    return call


from ._resource import ResourceManager
from ._process import  RestrictedWorker, RestrictedWorkerByJoin
from ._cache import CacheManager



import random

from autogoal.sampling import Sampler


class Grammar:
    def __init__(self, start):
        self._start = start

    def sample(self, *, max_iterations: int = 100, sampler: Sampler = None):
        if sampler is None:
            sampler = Sampler()

        return self._sample(
            symbol=self._start, max_iterations=max_iterations, sampler=sampler
        )

    def __call__(self, sampler: Sampler = None):
        return self.sample(sampler=sampler)

    def _sample(self, symbol, max_iterations, sampler):
        raise NotImplementedError()



import math
import random
import statistics
import pickle
import numpy as np

from typing import Dict, List, Sequence
import abc

from autogoal.utils import nice_repr


class Sampler:
    """
    Provides methods to obtain random samples with various distributions.

    Can receive a `random_state` to guarantee the same values are obtained
    in two different instantiations.
    """

    def __init__(self, *, random_state: int = None):
        self.rand = random.Random(random_state)

    def choice(self, options, handle=None):
        """
        Returns one of the options.

        ##### Examples

        ```python
        >>> sampler = Sampler(random_state=0)
        >>> [sampler.choice(['A', 'B', 'C']) for _ in range(5)]
        ['B', 'B', 'A', 'B', 'C']

        ```
        """
        return self.categorical(options, handle=handle)

    def distribution(self, name: str, handle=None, **kwargs):
        """
        Shortcut function for generating from a distribution,
        either `discrete`, `continuous`, `boolean` or `categorical`.
        """
        try:
            return getattr(self, name)(handle=handle, **kwargs)
        except AttributeError:
            raise ValueError("Unrecognized distribution name: %s" % name)

    def discrete(self, min=0, max=10, handle=None):
        """
        Returns a discrete value between `min` and `max`.

        ##### Examples

        ```python
        >>> sampler = Sampler(random_state=0)
        >>> [sampler.discrete(0, 10) for _ in range(10)]
        [6, 6, 0, 4, 8, 7, 6, 4, 7, 5]

        ```
        """
        return self.rand.randint(min, max)

    def continuous(self, min=0, max=1, handle=None):
        """
        Returns a continuous value between `min` and `max`.

        ##### Examples

        ```python
        >>> sampler = Sampler(random_state=0)
        >>> [round(sampler.continuous(0, 10), 2) for _ in range(10)]
        [8.44, 7.58, 4.21, 2.59, 5.11, 4.05, 7.84, 3.03, 4.77, 5.83]

        ```
        """
        return self.rand.uniform(min, max)

    def boolean(self, handle=None):
        """
        Returns a boolean value.

        ##### Examples

        ```python
        >>> sampler = Sampler(random_state=0)
        >>> [sampler.boolean() for _ in range(10)]
        [False, False, True, True, False, True, False, True, True, False]

        ```
        """
        return self.rand.uniform(0, 1) < 0.5

    def categorical(self, options, handle=None):
        """
        Returns one of the options.

        The difference between `choice` and `categorical` is evident in more specialized
        classes of `Sampler`. In the default implementation, their behavior is exactly the same.

        ##### Examples

        ```python
        >>> sampler = Sampler(random_state=0)
        >>> [sampler.categorical(['A', 'B', 'C']) for _ in range(5)]
        ['B', 'B', 'A', 'B', 'C']

        ```
        """
        return self.rand.choice(options)


class ModelSampler(Sampler):
    """
    A sampler that builds and uses an internal probabilistic model to generate
    values with a non-uniform probability.

    For the model to work, the `handler` parameter in each sampling method
    must be suplied, otherwise it behaves exactly as the standard `Sampler`.
    """

    def __init__(self, model: Dict = None, **kwargs):
        super().__init__(**kwargs)
        self._model: Dict = {} if model is None else model
        self._updates: Dict = {}

    @property
    def model(self):
        return self._model

    @property
    def updates(self):
        return self._updates

    def _get_model_params(self, handle, default):
        if handle in self._model:
            return self._model[handle]
        else:
            self._model[handle] = default
            return default

    def _register_update(self, handle, result):
        if handle not in self._updates:
            self._updates[handle] = []

        self._updates[handle].append(result)
        return result

    def _clamp(self, x, a, b):
        if x < a:
            return a
        if x > b:
            return b
        return x

    def choice(self, options, handle=None):
        if handle is not None:
            return self.categorical(options, handle)

        weights = [
            self._get_model_params(option, UnormalizedWeightParam(value=1))
            for option in options
        ]
        idx = self.rand.choices(
            range(len(options)), weights=[w.value for w in weights], k=1
        )[0]
        option = options[idx]
        self._register_update(option, 1)
        return option

    def discrete(self, min=0, max=10, handle=None):
        if handle is None:
            return super().discrete(min, max, handle)

        params = self._get_model_params(
            handle, MeanDevParam(mean=(min + max) / 2, dev=(max - min))
        )
        value = self._clamp(int(self.rand.gauss(params.mean, params.dev)), min, max)
        return self._register_update(handle, value)

    def continuous(self, min=0, max=1, handle=None):
        if handle is None:
            return super().continuous(min, max, handle)

        params = self._get_model_params(
            handle, MeanDevParam(mean=(min + max) / 2, dev=(max - min))
        )
        value = self._clamp(self.rand.gauss(params.mean, params.dev), min, max)
        return self._register_update(handle, value)

    def boolean(self, handle=None):
        if handle is None:
            return super().boolean(handle)

        params = self._get_model_params(handle, WeightParam(value=0.5))
        value = self.rand.uniform(0, 1) < params.value
        return self._register_update(handle, value)

    def categorical(self, options, handle=None):
        if handle is None:
            return super().categorical(options, handle)

        params = self._get_model_params(
            handle, DistributionParam(weights=[1 for _ in options])
        )
        idx = self.rand.choices(range(len(options)), weights=params.weights, k=1)[0]
        return options[self._register_update(handle, idx)]


class ReplaySampler:
    """
    A sampler that records the generated values and then can replay the
    same outputs in the same order.

    One of the most interesting use cases for `ReplaySampler` is in conjunction with context free
    or graph grammars, for generating complex objects.
    You can pass a sampler wrapped in a `ReplaySampler` during generation, and then
    reuse it later for generating the same object or graph.

    ##### Examples

    First, instantiate a `ReplaySampler` with an internal `Sampler` instance and
    use it normally.

    ```python
    >>> sampler = ReplaySampler(Sampler(random_state=0))
    >>> [sampler.discrete(0,10) for _ in range(10)]
    [6, 6, 0, 4, 8, 7, 6, 4, 7, 5]

    ```

    Then call the `replay` method and reuse the same values.
    `replay()` returns the same instance, to enable chaining method calls.

    ```python
    >>> sampler.replay()
    <autogoal.sampling.ReplaySampler object at ...>
    >>> [sampler.discrete(0,10) for _ in range(5)]
    [6, 6, 0, 4, 8]
    >>> [sampler.discrete(0,10) for _ in range(5)]
    [7, 6, 4, 7, 5]

    ```

    If you try to use it in a different way as originally, it will complain.

    ```python
    >>> sampler.replay().discrete(0,5)
    Traceback (most recent call last):
        ...
    TypeError: Invalid invocation of `discrete` with `args=(0, 5)`, replay history says args='(0, 10)'.

    >>> sampler.replay().boolean()
    Traceback (most recent call last):
        ...
    TypeError: Invalid invocation of `boolean`, replay history says discrete comes next.

    ```
    """

    RECORD = "record"
    REPLAY = "replay"

    def __init__(self, sampler):
        self.sampler = sampler
        self._mode = ReplaySampler.RECORD
        self._history = []
        self._current_history = []

    def _run(self, method, *args, **kwargs):
        if self._mode == ReplaySampler.RECORD:
            result = getattr(self.sampler, method)(*args, **kwargs)
            self._history.append(
                dict(method=method, args=repr(args), kwargs=repr(kwargs), result=result)
            )

            return result

        elif self._mode == ReplaySampler.REPLAY:
            if not self._current_history:
                raise TypeError(
                    f"Invalid invocation of `{method}`, replay history is empty. Maybe you forgot to call `replay`?"
                )

            top = self._current_history[0]

            if top["method"] != method:
                raise TypeError(
                    f"Invalid invocation of `{method}`, "
                    f"replay history says {top['method']} comes next."
                )

            if top["args"] != repr(args):
                raise TypeError(
                    f"Invalid invocation of `{method}` with `args={repr(args)}`, "
                    f"replay history says args={repr(top['args'])}."
                )

            if top["kwargs"] != repr(kwargs):
                raise TypeError(
                    f"Invalid invocation of `{method}` with `kwargs={repr(kwargs)}`, "
                    f"replay history says kwargs={repr(top['kwargs'])}."
                )

            self._current_history.pop(0)
            return top["result"]

    def replay(self) -> "ReplaySampler":
        self._mode = ReplaySampler.REPLAY
        self._current_history = list(self._history)
        return self

    def save(self, fp):
        """
        Saves the state of a `ReplaySampler` to a stream. It must be in replay mode.

        You are responsible for opening and closing the stream yourself.

        ##### Examples

        In this example we create a sampler, and save its state into a `StringIO`
        stream to be able to see what's being saved.

        ```python
        >>> sampler = ReplaySampler(Sampler(random_state=0))
        >>> [sampler.discrete(0, 10) for _ in range(3)]
        [6, 6, 0]

        >>> import io
        >>> fp = io.BytesIO()
        >>> sampler.replay().save(fp)
        >>> len(fp.getvalue())
        183

        ```
        """
        if self._mode != ReplaySampler.REPLAY:
            raise TypeError(
                "A sampler must be in replay mode, i.e., call the `replay()` method."
            )

        pickle.Pickler(fp).dump(self._history)

    @staticmethod
    def load(fp) -> "ReplaySampler":
        """
        Creates a `ReplaySampler` from a stream and returns it already in
        replay mode.

        You are responsible for opening and closing the stream yourself.

        ##### Examples

        ```python
        >>> sampler = ReplaySampler(Sampler(random_state=1))
        >>> [sampler.discrete(0, 10) for _ in range(10)]
        [2, 9, 1, 4, 1, 7, 7, 7, 10, 6]

        >>> import io
        >>> fp = io.BytesIO()
        >>> sampler.replay().save(fp)
        >>> fp.seek(0)
        0
        >>> other_sampler = ReplaySampler.load(fp)
        >>> [other_sampler.discrete(0, 10) for _ in range(5)]
        [2, 9, 1, 4, 1]
        >>> [other_sampler.discrete(0, 10) for _ in range(5)]
        [7, 7, 7, 10, 6]

        """
        history = pickle.Unpickler(fp).load()
        sampler = ReplaySampler(None)
        sampler._history = history
        return sampler.replay()

    def choice(self, *args, **kwargs):
        return self._run("choice", *args, **kwargs)

    def distribution(self, *args, **kwargs):
        return self._run("distribution", *args, **kwargs)

    def discrete(self, *args, **kwargs):
        return self._run("discrete", *args, **kwargs)

    def continuous(self, *args, **kwargs):
        return self._run("continuous", *args, **kwargs)

    def boolean(self, *args, **kwargs):
        return self._run("boolean", *args, **kwargs)

    def categorical(self, *args, **kwargs):
        return self._run("categorical", *args, **kwargs)

    def __getattr__(self, attr):
        return getattr(self.sampler, attr)


class ModelParam(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def update(self, alpha: float, updates) -> "ModelParam":
        pass


@nice_repr
class UnormalizedWeightParam(ModelParam):
    def __init__(self, value):
        self.value = value

    def update(self, alpha: float, updates: list) -> "UnormalizedWeightParam":
        return UnormalizedWeightParam(self.value + alpha * sum(updates))

    def weighted(self, solutions):
        result = 0
        for s, w in solutions:
            result += s * w

        return UnormalizedWeightParam(1 + result)


@nice_repr
class DistributionParam(ModelParam):
    def __init__(self, weights):
        total = sum(weights) or 1
        self.weights = [w / total for w in weights]

    def update(self, alpha: float, updates: list) -> "DistributionParam":
        weights = list(self.weights)

        for i in updates:
            weights[i] += alpha

        return DistributionParam(weights)

    def weighted(self, solutions):
        weights = [1] * len(self.weights)

        for s, w in solutions:
            weights[s] += w

        return DistributionParam(weights)


@nice_repr
class MeanDevParam(ModelParam):
    def __init__(self, mean, dev, *, initial_params=None):
        self.mean = mean
        self.dev = dev

        if initial_params is None:
            self.initial_params = (mean, dev)
        else:
            self.initial_params = initial_params

    def update(self, alpha: float, updates) -> "MeanDevParam":
        new_mean = statistics.mean(updates)
        new_dev = statistics.stdev(updates, new_mean) if len(updates) > 1 else 0

        return MeanDevParam(
            mean=self.mean * (1 - alpha) + new_mean * alpha,
            dev=self.dev * (1 - alpha) + new_dev * alpha,
            initial_params=self.initial_params,
        )

    def weighted(self, solutions):
        values = np.asarray(
            [s for s, w in solutions]
            + [
                self.initial_params[0] - 2 * self.initial_params[1],
                self.initial_params[0] + 2 * self.initial_params[1],
            ]
        )
        weights = np.asarray([w for s, w in solutions] + [1, 1])

        average = np.average(values, weights=weights)
        variance = np.average((values - average) ** 2, weights=weights)

        return MeanDevParam(average, math.sqrt(variance), initial_params=self.initial_params)


@nice_repr
class WeightParam(ModelParam):
    def __init__(self, value):
        self.value = value

    def update(self, alpha: float, updates) -> "WeightParam":
        new_value = statistics.mean(updates)
        return WeightParam(value=self.value * (1 - alpha) + new_value * alpha)

    def weighted(self, solutions):
        values = np.asarray([s for s, w in solutions] + [0, 1])
        weights = np.asarray([w for s, w in solutions] + [1, 1])

        return WeightParam(np.average(values, weights=weights))


def update_model(model, updates, alpha: float = 1):
    new_model = {}

    for handle, params in model.items():
        upd = updates.get(handle)

        if upd is None:
            new_model[handle] = params
        else:
            new_model[handle] = params.update(alpha, upd)

    return new_model


def _argsort(l):
    # taken from https://stackoverflow.com/questions/6422700
    return sorted(range(len(l)), key=l.__getitem__)


def best_indices(values: List, k: int = 1, maximize: bool = False) -> List[int]:
    """
    Computes the `k` best indices from values, i.e., the indices of the values
    that are the top minimum (or maximum).

    ##### Parameters

    * `values: List`: Values to compare, must be a sortable type (e.g., `int`, `float`, ...).
    * `k: int`: Number of indices to calculate. Defaults to `1`.
    * `maximize: bool`: Whether to compute the maximum or minimum values. Defaults to `False`, i.e., minimize by default.

    ##### Returns:

    * `indices: List[int]`: list of the indices that correspond to maximum (or minimum) values in `values`.

    ##### Examples:

    ```python
    >>> best_indices([.33, 0.12, 0.55, 0.09], k=2)
    [1, 3]

    >>> best_indices([.33, 0.12, 0.55, 0.09], k=3, maximize=True)
    [0, 1, 2]

    >>> best_indices([.33, 0.12, 0.55, 0.09])
    [3]

    ```

    !!! note
        Note that indices are returned in their original order, **not** in the order in which
        the values would be sorted themselves.
    """
    indices = _argsort(_argsort(values))

    if maximize:
        threshold = len(values) - k
        return [i for i in range(len(values)) if indices[i] >= threshold]
    else:
        threshold = k
        return [i for i in range(len(values)) if indices[i] < threshold]


def merge_updates(*updates: Sequence[Dict]) -> Dict:
    """
    Merges a bunch of update dicts from `ModelSampler`
    into a single dictionary.

    ##### Parameters:

    * `updates: Sequence[Dict]`: Sequence of update dictionaries obtained
      from calling `ModelSampler.updates`.

    ##### Returns:

    * `update: Dict`: A single dictionary with the combined (appended) updates.

    ##### Examples:

    ```python
    >>> up1 = {'a': [1]}
    >>> up2 = {'b': [2,3]}
    >>> up3 = {'a': [4]}
    >>> merge_updates(up1, up2, up3)
    {'a': [1, 4], 'b': [2, 3]}

    ```
    """
    result = {}

    for upd in updates:
        for key, value in upd.items():
            if not key in result:
                result[key] = []

            result[key].extend(value)

    return result



from autogoal.datasets.ehealthkd20._utils import Collection, Sentence


def to_biluov(tokensxsentence, entitiesxsentence):
    labelsxsentence = []
    for tokens, entities in zip(tokensxsentence, entitiesxsentence):
        offset = 0
        labels = []
        for token in tokens:
            # Recently found that (token.idx, token.idx + len(token)) is the span
            matches = find_match(offset, offset + len(token.text), entities)
            tag = select_tag(matches)
            labels.append(tag)
            offset += len(token.text_with_ws)
        labelsxsentence.append(labels)

    return labelsxsentence#, "BILUOV"


def find_match(start, end, entities):
    def match(other):
        return other[0] <= start and end <= other[1]

    matches = []
    for spans in entities:

        # UNIT
        if len(spans) == 1:
            if match(spans[0]):
                matches.append((spans[0], "U"))
            continue

        # BEGIN
        begin, *tail = spans
        if match(begin):
            matches.append((begin, "B"))
            continue

        # LAST
        *body, last = tail
        if match(last):
            matches.append((last, "L"))
            continue

        # INNER
        for inner in body:
            if match(inner):
                matches.append((inner, "I"))
                break

    return matches


def select_tag(matches):
    if not matches:
        return "O"
    if len(matches) == 1:
        return matches[0][1]
    tags = [tag for _, tag in matches]
    return "U" if ("U" in tags and not "B" in tags and not "L" in tags) else "V"


def make_sentence(doc, bilouv, labels) -> Sentence:
    sentence = Sentence(doc.text)

    logger.debug(f"[make_sentence]: doc.text={doc.text}")
    logger.debug(f"[make_sentence]: bilouv={bilouv}")

    labels = set(l[2:] for l in labels if l != 'O')
    
    for label in labels:
        specific_bilouv = []

        for tag in bilouv:
            if tag.endswith(label):
                tag = tag[0]
                specific_bilouv.append(tag[0])
            else:
                specific_bilouv.append('O')

        logger.debug(f"[make_sentence]: label={label} specific_bilouv={specific_bilouv}")

        spans = from_biluov(specific_bilouv, doc, spans=True)
        sentence.keyphrases.extend(Keyphrase(sentence, label, i, sp) for i,sp in enumerate(spans))

    return sentence


def from_biluov(biluov, sentence, *, spans=False, drop_remaining=[]):
    """
    >>> from_biluov(list('BBULL'), 'A B C D E'.split())
    [['C'], ['B', 'D'], ['A', 'E']]
    """

    entities = [x for x in discontinuous_match(biluov, sentence)]

    for i, (tag, word) in enumerate(zip(biluov, sentence)):
        if tag == "U":
            entities.append([word])
            biluov[i] = "O"
        elif tag == "V":
            biluov[i] = "I"

    # only BILO is left!!!
    changed = True
    while changed:
        changed = False
        one_shot = enumerate(zip(biluov, sentence))
        try:
            i, (tag, word) = next(one_shot)
            while True:
                if tag != "B":
                    i, (tag, word) = next(one_shot)
                    continue

                on_build = [(word, i)]

                i, (tag, word) = next(one_shot)
                while tag in ("O", "I"):
                    if tag == "I":
                        on_build.append(word)
                    i, (tag, word) = next(one_shot)

                if tag == "L":
                    entities.append([x for x, _ in on_build] + [word])
                    for _, j in on_build:
                        biluov[j] = "O"
                    biluov[i] = "O"
                    on_build.clear()
                    changed = True
        except StopIteration:
            pass

    for i, (tag, word) in enumerate(zip(biluov, sentence)):
        if tag != "O" and tag not in drop_remaining:
            entities.append([word])

    return (
        entities
        if not spans
        else [[(t.idx, t.idx + len(t)) for t in tokens] for tokens in entities]
    )


def discontinuous_match(biluov, sentence):
    """
    >>> discontinuous_match(['B','V','L'],['la', 'enfermedad', 'renal'])
    [['la', 'enfermedad', 'renal'], ['enfermedad']]
    >>> discontinuous_match(['O','V','I','L','O','I','L'],['el','cancer','de','pulmon','y','de','mama'])
    [['cancer', 'de', 'pulmon'], ['cancer', 'de', 'mama']]
    >>> discontinuous_match(['B','O','B','V'],['tejidos','y','organos','humanos'])
    [['organos', 'humanos'], ['tejidos', 'humanos']]
    >>> discontinuous_match(['O','V','I','L','O','I','L','O','B','O','B','V'], ['el','cancer','de','pulmon','y','de','mama','y','tejidos','y','organos','humanos'])
    [['cancer', 'de', 'pulmon'], ['cancer', 'de', 'mama'], ['organos', 'humanos'], ['tejidos', 'humanos']]
    >>> discontinuous_match(list('BBULL'), 'A B C D E'.split())
    []
    """
    entities = []
    for i, tag in enumerate(biluov):
        if tag != "V":
            continue
        for entity_ids in _full_overlap(biluov, list(range(len(sentence))), i):
            entity = []
            for idx in entity_ids:
                entity.append(sentence[idx])
                biluov[idx] = "O"
            entities.append(entity)
    return entities


def _full_overlap(biluov, sentence, index, product=False):
    """
    INDEX TAG MUST BE 'V'
    >>> _full_overlap(['B','V','L'], list(range(3)), 1)
    [[0, 1, 2], [1]]
    >>> _full_overlap(['B','V','V','L'], list(range(4)), 1)
    [[0, 1, 2, 3], [1, 2]]
    >>> _full_overlap(['B','V','V','L'], list(range(4)), 2)
    [[0, 1, 2, 3], [1, 2]]
    >>> _full_overlap(['B','V','V','V','L'], list(range(5)), 1)
    [[0, 1, 2, 3, 4], [1, 2, 3]]
    >>> _full_overlap(['B','V','V','V','L'], list(range(5)), 2)
    [[0, 1, 2, 3, 4], [1, 2, 3]]
    >>> _full_overlap(['B','V','V','V','L'], list(range(5)), 3)
    [[0, 1, 2, 3, 4], [1, 2, 3]]
    >>> _full_overlap(['B','B','V','L','L'], list(range(5)), 2)
    [[1, 2, 3], [0, 2, 4]]
    >>> _full_overlap(['B','I','B','O','V','I','L','O','L'], list(range(9)), 4)
    [[2, 4, 5, 6], [0, 1, 4, 8]]
    >>> _full_overlap(['B','I','B','O','V','I','L','O','L'], list(range(9)), 4, True)
    [[2, 4, 5, 6], [2, 4, 8], [0, 1, 4, 5, 6], [0, 1, 4, 8]]
    >>> _full_overlap(['0','0','V','L'], list(range(4)), 2)
    [[2, 3], [2]]
    >>> _full_overlap(['V','L'], list(range(2)), 0)
    [[0, 1], [0]]
    >>> _full_overlap(['B','V','O','O'], list(range(4)), 1)
    [[0, 1], [1]]
    >>> _full_overlap(['B','V'], list(range(2)), 1)
    [[0, 1], [1]]
    >>> _full_overlap(['0','0','V','O','O'], list(range(5)), 2)
    []
    """

    left = _right_to_left_overlap(biluov[: index + 1], sentence[: index + 1])
    right = _left_to_right_overlap(biluov[index:], sentence[index:])

    full = []
    if product:
        for l in left:
            for r in right:
                new = l + r[1:] if len(l) > len(r) else l[:-1] + r
                full.append(new)
    else:
        for l, r in itt.zip_longest(left, right, fillvalue=[]):
            new = l + r[1:] if len(l) > len(r) else l[:-1] + r
            full.append(new)
    return full


def _left_to_right_overlap(biluov, sentence):
    """
    LEFTMOST TAG MUST BE 'V'
    >>> _left_to_right_overlap(['V', 'V', 'O', 'V', 'I', 'L', 'O', 'I', 'L'], range(9))
    [[0, 1, 3, 4, 5], [0, 1, 3, 7, 8]]
    >>> _left_to_right_overlap(['V', 'O', 'V', 'O'], range(4))
    []
    >>> _left_to_right_overlap(['V', 'O', 'V', 'O', 'L'], range(5))
    [[0, 2, 4], [0, 2]]
    >>> _left_to_right_overlap(['V', 'O', 'V', 'O', 'L', 'O', 'L'], range(8))
    [[0, 2, 4], [0, 2, 6]]
    >>> _left_to_right_overlap(['V', 'O', 'V', 'O', 'L', 'I', 'L', 'V', 'L'], range(9))
    [[0, 2, 4], [0, 2, 5, 6]]
    """
    return _build_overlap(biluov, sentence, "L")


def _right_to_left_overlap(biluov, sentence):
    """
    RIGHTMOST TAG MUST BE 'V'
    >>> _right_to_left_overlap(['B', 'I', 'O', 'B', 'I', 'V', 'O', 'V', 'V'], range(9))
    [[3, 4, 5, 7, 8], [0, 1, 5, 7, 8]]
    >>> _right_to_left_overlap(['O', 'V', 'O', 'V'], range(4))
    []
    >>> _right_to_left_overlap(['B', 'O', 'V', 'O', 'V'], range(5))
    [[0, 2, 4], [2, 4]]
    >>> _right_to_left_overlap(['B', 'O', 'B', 'O', 'V', 'O', 'V'], range(7))
    [[2, 4, 6], [0, 4, 6]]
    >>> _right_to_left_overlap(['B', 'V', 'B', 'I', 'B', 'O', 'V', 'O', 'V'], range(9))
    [[4, 6, 8], [2, 3, 6, 8]]
    """
    inverse = _build_overlap(reversed(biluov), reversed(sentence), "B")
    for x in inverse:
        x.reverse()
    return inverse


def _build_overlap(biluov, sentence, finisher):
    """
    LEFTMOST TAG MUST BE 'V'
    """

    one_shot = zip(biluov, sentence)
    tag, word = next(one_shot)

    prefix = []
    complete = []

    try:
        while tag in ("V", "O"):
            if tag == "V":
                prefix.append(word)
            tag, word = next(one_shot)

        on_build = []
        while tag in ("O", "I", "U", finisher):
            if tag == "I":
                on_build.append(word)
            elif tag == finisher:
                complete.append(prefix + on_build + [word])
                on_build.clear()
            elif tag == "U":
                complete.append([word])
            tag, word = next(one_shot)
    except StopIteration:
        pass

    if len(complete) == 1:
        complete.append(prefix)

    return complete



# coding: utf-8

import os
import numpy as np
from scipy import sparse as sp
from autogoal.datasets import datapath, download

def load():
    """
    Loads train and valid datasets from [DOROTHEA uci dataset](https://archive.ics.uci.edu/ml/datasets/dorothea).

    ##### Examples

    ```python
    >>> X_train, y_train, X_valid, y_valid = load()
    >>> X_train.shape, X_valid.shape
    ((800, 100000), (350, 100000))
    >>> len(y_train), len(y_valid)
    (800, 350)

    ```
    """

    download("dorothea")

    train_data = open(datapath('dorothea')/ "dorothea_train.data", "r")
    train_labels = open(datapath('dorothea')/ "dorothea_train.labels", "r")
    valid_data = open(datapath('dorothea')/ "dorothea_valid.data", "r")
    valid_labels = open(datapath('dorothea')/ "dorothea_valid.labels", "r")


    Xtrain = sp.lil_matrix((800, 100000), dtype=int)
    ytrain = []
    Xvalid = sp.lil_matrix((350, 100000), dtype=int)
    yvalid = []

    for row, line in enumerate(train_data):
        for col in line.split():
            Xtrain[row, int(col)-1] = 1

    for row, line in enumerate(valid_data):
        for col in line.split():
            Xvalid[row, int(col)-1] = 1

    for line in train_labels:
        ytrain.append(int(line))

    for line in valid_labels:
        yvalid.append(int(line))

    return Xtrain.tocsr(), np.asarray(ytrain), Xvalid.tocsr(), np.asarray(yvalid)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autogoal',
 'autogoal.contrib',
 'autogoal.contrib.gensim',
 'autogoal.contrib.keras',
 'autogoal.contrib.nltk',
 'autogoal.contrib.regex',
 'autogoal.contrib.sklearn',
 'autogoal.contrib.spacy',
 'autogoal.contrib.streamlit',
 'autogoal.contrib.telegram',
 'autogoal.contrib.torch',
 'autogoal.contrib.wikipedia',
 'autogoal.datasets',
 'autogoal.datasets.ehealthkd20',
 'autogoal.grammar',
 'autogoal.kb',
 'autogoal.ml',
 'autogoal.sampling',
 'autogoal.search',
 'autogoal.utils']

package_data = \
{'': ['*'], 'autogoal.datasets': ['data/.gitignore']}

install_requires = \
['black>=19.10b0,<20.0',
 'enlighten>=1.4.0,<2.0.0',
 'networkx>=2.4,<3.0',
 'numpy>=1.19.2,<2.0.0',
 'psutil>=5.6.7,<6.0.0',
 'pydot>=1.4.1,<2.0.0',
 'pyyaml>=5.2,<6.0',
 'scipy>=1.5.2,<2.0.0',
 'termcolor>=1.1.0,<2.0.0',
 'toml>=0.10.0,<0.11.0',
 'tqdm>=4.50.2,<5.0.0',
 'typer>=0.3.2,<0.4.0']

extras_require = \
{'contrib': ['gensim>=3.8.1,<4.0.0',
             'jupyterlab>=1.2.4,<2.0.0',
             'keras>=2.3.1,<3.0.0',
             'nltk>=3.4.5,<4.0.0',
             'nx_altair>=0.1.4,<0.2.0',
             'python-telegram-bot>=12.4.2,<13.0.0',
             'scikit-learn>=0.22,<0.23',
             'seqlearn>=0.2,<0.3',
             'sklearn_crfsuite>=0.3.6,<0.4.0',
             'spacy>=2.2.3,<3.0.0',
             'streamlit>=0.59.0,<0.60.0',
             'transformers>=2.3.0,<3.0.0',
             'wikipedia>=1.4.0,<2.0.0'],
 'dev': ['codecov>=2.0.15,<3.0.0',
         'markdown-include>=0.5.1,<0.6.0',
         'mkdocs>=1.0.4,<2.0.0',
         'mkdocs-material>=4.6.0,<5.0.0',
         'mypy>=0.761,<0.762',
         'pylint>=2.4.4,<3.0.0',
         'pytest>=5.3.2,<6.0.0',
         'pytest-cov>=2.8.1,<3.0.0'],
 'gensim': ['gensim>=3.8.1,<4.0.0'],
 'keras': ['keras>=2.3.1,<3.0.0'],
 'nltk': ['nltk>=3.4.5,<4.0.0'],
 'sklearn': ['scikit-learn>=0.22,<0.23',
             'seqlearn>=0.2,<0.3',
             'sklearn_crfsuite>=0.3.6,<0.4.0'],
 'spacy': ['spacy>=2.2.3,<3.0.0'],
 'streamlit': ['nx_altair>=0.1.4,<0.2.0', 'streamlit>=0.59.0,<0.60.0'],
 'telegram': ['python-telegram-bot>=12.4.2,<13.0.0'],
 'wikipedia': ['wikipedia>=1.4.0,<2.0.0']}

setup_kwargs = {
    'name': 'autogoal',
    'version': '0.3.2',
    'description': 'Automatic Generation Optimization And Learning',
    'long_description': '![AutoGOAL Logo](https://autogoal.github.io/autogoal-banner.png)\n\n[<img alt="PyPI" src="https://img.shields.io/pypi/v/autogoal">](https://pypi.org/project/autogoal/) [<img alt="PyPI - Python Version" src="https://img.shields.io/pypi/pyversions/autogoal">](https://pypi.org/project/autogoal/) [<img alt="PyPI - License" src="https://img.shields.io/pypi/l/autogoal">](https://autogoal.github.io/contributing) [<img alt="GitHub stars" src="https://img.shields.io/github/stars/autogoal/autogoal?style=social">](https://github.com/autogoal/autogoal/stargazers) [<img alt="Twitter Follow" src="https://img.shields.io/twitter/follow/auto_goal?label=Followers&style=social">](https://twitter.com/auto_goal)\n\n[<img alt="GitHub Workflow Status (branch)" src="https://img.shields.io/github/workflow/status/autogoal/autogoal/CI/main?label=unit tests&logo=github">](https://github.com/autogoal/autogoal/actions)\n[<img src="https://codecov.io/gh/autogoal/autogoal/branch/main/graph/badge.svg" />](https://codecov.io/gh/autogoal/autogoal/)\n[<img alt="Docker Cloud Build Status" src="https://img.shields.io/docker/cloud/build/autogoal/autogoal">](https://hub.docker.com/r/autogoal/autogoal)\n[<img alt="Docker Image Size (CPU)" src="https://img.shields.io/docker/image-size/autogoal/autogoal/latest">](https://hub.docker.com/r/autogoal/autogoal)\n[<img alt="Docker Pulls" src="https://img.shields.io/docker/pulls/autogoal/autogoal">](https://hub.docker.com/r/autogoal/autogoal)\n\n# AutoGOAL\n\n> Automatic Generation, Optimization And Artificial Learning\n\nAutoGOAL is a Python library for automatically finding the best way to solve a given task.\nIt has been designed mainly for _Automated Machine Learning_ (aka [AutoML](https://www.automl.org))\nbut it can be used in any scenario where you have several possible ways to solve a given task.\n\nTechnically speaking, AutoGOAL is a framework for program synthesis, i.e., finding the best program to solve\na given problem, provided that the user can describe the space of all possible programs.\nAutoGOAL provides a set of low-level components to define different spaces and efficiently search in them.\nIn the specific context of machine learning, AutoGOAL also provides high-level components that can be used as a black-box in almost any type of problem and dataset format.\n\n## Quickstart\n\nAutoGOAL is first and foremost a framework for Automated Machine Learning.\nAs such, it comes pre-packaged with hundreds of low-level machine learning\nalgorithms that can be automatically assembled into pipelines for different problems.\n\nThe core of this functionality lies in the [`AutoML`](https://autogoal.github.io/api/autogoal.ml#automl) class.\n\nTo illustrate the simplicity of its use we will load a dataset and run an automatic classifier in it.\n\n```python\nfrom autogoal.datasets import cars\nfrom autogoal.ml import AutoML\n\nX, y = cars.load()\nautoml = AutoML()\nautoml.fit(X, y)\n```\n\nSensible defaults are defined for each of the many parameters of `AutoML`.\nMake sure to [read the documentation](https://autogoal.github.io/guide/) for more information.\n\n## Installation\n\nInstallation is very simple:\n\n    pip install autogoal\n\nHowever, `autogoal` comes with a bunch of optional dependencies. You can install them all with:\n\n    pip install autogoal[contrib]\n\nTo fine-pick which dependencies you want, read the [dependencies section](https://autogoal.github.io/dependencies/).\n\n### Using Docker \n\nThe easiest way to get AutoGOAL up and running with all the dependencies is to pull the development Docker image, which is somewhat big:\n\n    docker pull autogoal/autogoal\n\nInstructions for setting up Docker are available [here](https://www.docker.com/get-started).\n\nOnce you have the development image downloaded, you can fire up a console and use AutoGOAL interactively.\n\n![](https://autogoal.github.io/shell.svg)\n\n> **NOTE**: By installing through `pip` you will get the latest release version of AutoGOAL, while by installing through Docker, you will get the latest development version. The development version is mostly up-to-date with the `main` branch, hence it will probably contain more features, but also more bugs, than the release version.\n\n## CLI\n\nYou can use AutoGOAL directly from the CLI. To see options just type:\n\n    python -m autogoal\n\nRead more in the [CLI documentation](https://autogoal.github.io/cli).\n\n## Demo\n\nAn online demo app is available at [autogoal.github.io/demo](https://autogoal.github.io/demo).\nThis app showcases the main features of AutoGOAL in interactive case studies.\n\nTo run the demo locally, simply type:\n\n    docker run -p 8501:8501 autogoal/autogoal\n\nAnd navigate to [localhost:8501](http://localhost:8501).\n\n## Documentation\n\nThis documentation is available online at [autogoal.github.io](https://autogoal.github.io). Check the following sections:\n\n- [**User Guide**](https://autogoal.github.io/guide/): Step-by-step showcase of everything you need to know to use AuoGOAL.\n- [**Examples**](https://autogoal.github.io/examples/): The best way to learn how to use AutoGOAL by practice.\n- [**API**](https://autogoal.github.io/api/autogoal): Details about the public API for AutoGOAL.\n\nThe HTML version can be deployed offline by downloading the [AutoGOAL Docker image](https://hub.docker.com/autogoal/autogoal) and running:\n\n    docker run -p 8000:8000 autogoal/autogoal mkdocs serve -a 0.0.0.0:8000\n\nAnd navigating to [localhost:8000](http://localhost:8000).\n\n## Publications\n\nIf you use AutoGOAL in academic research, please cite the following paper:\n\n```bibtex\n@article{estevez2020general,\n  title={General-purpose hierarchical optimisation of machine learning pipelines with grammatical evolution},\n  author={Est{\\\'e}vez-Velarde, Suilan and Guti{\\\'e}rrez, Yoan and Almeida-Cruz, Yudivi{\\\'a}n and Montoyo, Andr{\\\'e}s},\n  journal={Information Sciences},\n  year={2020},\n  publisher={Elsevier},\n  doi={10.1016/j.ins.2020.07.035}\n}\n```\n\nThe technologies and theoretical results leading up to AutoGOAL have been presented at different venues:\n\n- [Optimizing Natural Language Processing Pipelines: Opinion Mining Case Study](https://link.springer.com/chapter/10.1007/978-3-030-33904-3_15) marks the inception of the idea of using evolutionary optimization with a probabilistic search space for pipeline optimization.\n\n- [AutoML Strategy Based on Grammatical Evolution: A Case Study about Knowledge Discovery from Text](https://www.aclweb.org/anthology/P19-1428/) applied probabilistic grammatical evolution with a custom-made grammar in the context of entity recognition in medical text.\n\n- [General-purpose Hierarchical Optimisation of Machine Learning Pipelines with Grammatical Evolution](https://doi.org/10.1016/j.ins.2020.07.035) presents a more uniform framework with different grammars in different problems, from tabular datasets to natural language processing.\n\n- [Solving Heterogeneous AutoML Problems with AutoGOAL](https://www.automl.org/wp-content/uploads/2020/07/AutoML_2020_paper_20.pdf) is the first actual description of AutoGOAL as a framework, unifying the ideas presented in the previous papers.\n\n## Contribution\n\nCode is licensed under MIT. Read the details in the [collaboration section](https://autogoal.github.io/contributing).\n',
    'author': 'Suilan Estevez-Velarde',
    'author_email': 'suilanestevez@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://autogoal.github.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
AutoGOAL is a Python framework for the automatic optimization, generation and learning of software pipelines.

A software pipeline is defined, for the purpose of AutoGOAL, as any software component, whether a class hierarchy,
a set of functions, or any combination thereof, that work together to solve a specific problem.
With AutoGOAL you can define a pipeline in many different ways, such that certain parts of it are configurable or
tunable, and then use search algorithms to find the best way to tune or configure it for a given problem.
"""

from autogoal import contrib
from autogoal import datasets
from autogoal import grammar
from autogoal import kb
from autogoal import ml
from autogoal import sampling
from autogoal import search
from autogoal import utils

from autogoal.utils._helpers import optimize



import shutil
import json
import requests
import os
from tqdm import tqdm

from pathlib import Path

DATASETS_METADATA = (
    "https://raw.githubusercontent.com/autogoal/datasets/master/datasets.json"
)


def datapath(path: str) -> Path:
    """
    Returns a `Path` object that points to the dataset path
    where `path` is located.

    ##### Examples

    ```python
    >>> datapath("movie_reviews")
    PosixPath('/home/coder/autogoal/autogoal/datasets/data/movie_reviews')

    ```
    """
    return Path(__file__).parent / "data" / path


def pack(folder: str):
    filename = datapath(folder)
    rootdir = datapath(folder)
    shutil.make_archive(filename, "zip", root_dir=rootdir)


def unpack(zipfile: str):
    filename = datapath(zipfile)
    rootdir = datapath(zipfile[:-4])
    shutil.unpack_archive(filename, extract_dir=rootdir, format="zip")


def download(dataset: str, unpackit: bool = True):
    fname = f"{dataset}.zip"
    path = datapath(fname)

    if path.exists():
        return

    datasets = requests.get(DATASETS_METADATA).json()
    url = datasets[dataset]

    download_and_save(url, path, True)

    if unpackit:
        unpack(fname)


def download_and_save(url, path: Path, overwrite=False, data_length=None):
    stream = requests.get(url, stream=True)
    total_size = data_length or int(stream.headers.get('content-length', 0))

    if path.exists() and not overwrite:
        return False

    try:
        with path.open("wb") as f:
            with tqdm(total=total_size, unit='B', unit_scale=True, unit_divisor=1024) as pbar:
                for data in stream.iter_content(32*1024):
                    f.write(data)
                    pbar.update(len(data))

        return True
    except:
        path.unlink()
        raise



import networkx as nx
import random
import types

from typing import List, Dict

from autogoal.utils import nice_repr
from ._base import Grammar, Sampler


class Graph(nx.DiGraph):
    """
    Represents a graph in AutoGOAL that can be expanded with graph grammars.
    """

    def __init__(self, **attrs):
        super(Graph, self).__init__(**attrs)

    def build_order(self):
        """
        Returns a iterable of `(node, in_nodes)` in topological order.

        In detail, returns every `node` and the list of nodes that are incoming edges to `node`.
        This way, you can call a method on each node whose args are the
        previous nodes.
        """
        for node in nx.topological_sort(self):
            in_nodes = [u for u, v in self.in_edges(node)]
            yield (node, in_nodes)

    def apply(self, function):
        """
        Applies a function to all nodes in `build_order`'s order.

        The function receives the current three arguments:
            - The current `node` instance
            - A list of incoming node instances
            - A list of the result of the application of the
              function in the previous node instances.

        Returns the last of the values computed.
        """
        previous_values = {}
        final_values = []

        for node, in_nodes in self.build_order():
            in_values = [previous_values[n] for n in in_nodes]
            value = function(node, in_nodes, in_values)

            if not list(self.neighbors(node)):
                final_values.append(value)

            previous_values[node] = value

        return final_values

    def contains_any(self, *items):
        return any((node in items) for node in self)


def uniform_selection(items):
    return random.choice(items)


def first_selection(items):
    return items[0]


def default_initializer(cls, sampler=None):
    return cls()


# Holds a string to class map for automatically generated classes
_GENERATED_CLASSES: Dict[str, type] = {}


def _get_generated_class(name):
    if name in _GENERATED_CLASSES:
        return _GENERATED_CLASSES[name]

    clss = types.new_class(name)
    _GENERATED_CLASSES[name] = clss

    return clss


@nice_repr
class Production:
    def __init__(self, production_id, pattern, replacement, *, initializer=None):
        if not isinstance(pattern, Graph):
            obj = pattern
            pattern = Graph()
            pattern.add_node(obj)

        if not isinstance(replacement, GraphPattern):
            replacement = Node(replacement)

        self.production_id = production_id
        self.pattern = pattern
        self.replacement = replacement
        self._initializer = initializer or default_initializer

    def _matches(self, graph: Graph):
        # TODO: Generalizar a permitir cualquier tipo de grafo como patrón, no solo un nodo
        pattern_node = list(self.pattern.nodes)[0]

        for node in graph.nodes:
            if node.__class__ == pattern_node:
                yield node

    def match(self, graph: Graph):
        """
        Returns True if it finds a subgraph in `graph` that matches the pattern.
        NOTE: Right now only works if pattern is a single node.
        """
        for _ in self._matches(graph):
            return True

        return False

    def apply(self, graph: Graph, pattern_selection=uniform_selection) -> Graph:
        """
        Applies a production in a graph and returns the modified graph.
        """
        matches = list(self._matches(graph))
        node = pattern_selection(matches)

        in_edges = graph.in_edges(node)
        out_edges = graph.out_edges(node)

        in_nodes = [u for u, v in in_edges]
        out_nodes = [v for u, v in out_edges]

        graph.remove_node(node)
        self.replacement.build(
            graph, in_nodes=in_nodes, out_nodes=out_nodes, initializer=self._initializer
        )

        return graph


class GraphPattern:
    def build(
        self,
        graph: Graph,
        *,
        in_nodes=[],
        out_nodes=[],
        initializer=default_initializer,
    ):
        raise NotImplementedError()

    def _add_in_nodes(self, graph, in_nodes, node):
        for in_node in in_nodes:
            graph.add_edge(in_node, node)

    def _add_out_nodes(self, graph, out_nodes, node):
        for out_node in out_nodes:
            graph.add_edge(node, out_node)

    def make(self, *, initializer=default_initializer) -> Graph:
        graph = Graph()
        self.build(graph, in_nodes=[], out_nodes=[], initializer=initializer)
        return graph


class Node(GraphPattern):
    def __init__(self, cls):
        self.cls = _get_generated_class(cls) if isinstance(cls, str) else cls

    def build(
        self,
        graph: Graph,
        *,
        in_nodes=[],
        out_nodes=[],
        initializer=default_initializer,
    ):
        obj = initializer(self.cls)
        graph.add_node(obj)
        self._add_in_nodes(graph, in_nodes, obj)
        self._add_out_nodes(graph, out_nodes, obj)


class Path(GraphPattern):
    def __init__(self, *items):
        self.items = [
            (_get_generated_class(i) if isinstance(i, str) else i) for i in items
        ]

    def build(
        self,
        graph: Graph,
        *,
        in_nodes=[],
        out_nodes=[],
        initializer=default_initializer,
    ):
        items = [initializer(cls) for cls in self.items]
        graph.add_nodes_from(items)

        for i, j in zip(items, items[1:]):
            graph.add_edge(i, j)

        self._add_in_nodes(graph, in_nodes, items[0])
        self._add_out_nodes(graph, out_nodes, items[-1])


class Block(GraphPattern):
    def __init__(self, *items):
        self.items = [
            (_get_generated_class(i) if isinstance(i, str) else i) for i in items
        ]

    def build(
        self,
        graph: Graph,
        *,
        in_nodes=[],
        out_nodes=[],
        initializer=default_initializer,
    ):
        items = [initializer(cls) for cls in self.items]
        graph.add_nodes_from(items)

        for item in items:
            self._add_in_nodes(graph, in_nodes, item)
            self._add_out_nodes(graph, out_nodes, item)


class Epsilon(GraphPattern):
    def build(
        self,
        graph: Graph,
        *,
        in_nodes=[],
        out_nodes=[],
        initializer=default_initializer,
    ):
        for u in in_nodes:
            for v in out_nodes:
                graph.add_edge(u, v)


class GraphGrammar(Grammar):
    def __init__(self, start, *, initializer=None, non_terminals=None):
        if isinstance(start, str):
            start = Node(start)

        if isinstance(start, GraphPattern):
            start = start.make()

        super(GraphGrammar, self).__init__(start)
        self._productions: List[Production] = []
        self._non_terminals = set(non_terminals or [])
        self._initializer = initializer or default_initializer

    def add(self, pattern, replacement: GraphPattern, *, initializer=None, kwargs=None):
        if initializer is None:
            initializer = self._initializer

        if kwargs is not None:
            initializer = lambda cls: cls(**kwargs)

        if isinstance(pattern, str):
            pattern = _get_generated_class(pattern)
            self._non_terminals.add(pattern)

        self._productions.append(
            Production("production_%i" % len(self._productions), pattern, replacement, initializer=initializer)
        )

    def _sample(self, symbol: Graph, max_iterations: int, sampler: Sampler):
        if symbol is None:
            raise ValueError("`symbol` cannot be `None`")

        symbol = symbol.copy()

        while max_iterations > 0:
            valid_productions = [p for p in self._productions if p.match(symbol)]

            if self._non_terminals:
                non_terminal_productions = [
                    p
                    for p in valid_productions
                    if p.pattern.contains_any(*self._non_terminals)
                ]

                if non_terminal_productions:
                    valid_productions = non_terminal_productions

            if not valid_productions:
                return symbol

            production_ids = {p.production_id: p for p in valid_productions} 
            production = production_ids[sampler.choice(list(production_ids))]
            symbol = production.apply(symbol)

            max_iterations -= 1

        return symbol

    def __repr__(self):
        return repr(self._productions)


@nice_repr
class Start:
    __slots__ = ()
    __name__ = "Start"

    def __eq__(self, other):
        return isinstance(other, Start)

    def __hash__(self):
        return 0


@nice_repr
class End:
    __slots__ = ()
    __name__ = "End"

    def __eq__(self, other):
        return isinstance(other, End)

    def __hash__(self):
        return 0


class GraphSpace(Grammar):
    Start = Start()
    End = End()

    def __init__(self, graph: Graph, initializer=None):
        self.graph = graph
        self.initializer = initializer or default_initializer

    @property
    def _start(self):
        return [GraphSpace.Start]

    def _sample(self, symbol, max_iterations, sampler):
        path = symbol
        visited = set()

        while max_iterations > 0:
            last_node = path[-1]

            if last_node == GraphSpace.End:
                break

            # next_nodes = list(set(self.graph.neighbors(last_node)) - visited)
            next_nodes = list(self.graph.neighbors(last_node))

            if not next_nodes:
                raise ValueError("Cannot continue sampling. Graph is disconnected.")

            names_values = { x.__name__: x for x in next_nodes }

            next_node = names_values[sampler.choice(list(names_values))]
            path.append(next_node)
            visited.add(next_node)
        else:
            raise ValueError("Reached maximum iterations")

        return [self.initializer(node, sampler=sampler) for node in path[1:-1]]



import statistics
import abc

from typing import Mapping, Optional, Dict, List, Sequence
from autogoal.sampling import ModelSampler, best_indices, merge_updates, update_model
from ._base import SearchAlgorithm

import random

import pickle
import time

class PESearch(SearchAlgorithm):
    def __init__(
        self,
        *args,
        learning_factor: float = 0.05,
        selection: float = 0.2,
        epsilon_greed: float = 0.1,
        random_state: Optional[int] = None,
        name: str = None,
        save: bool = False,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._learning_factor = learning_factor
        self._selection = selection
        self._epsilon_greed = epsilon_greed
        self._model: Dict = {}
        self._random_states = random.Random(random_state)
        self._name = name or str(time.time())
        self._save = save

    def _start_generation(self):
        self._samplers = []

    def _build_sampler(self):
        if len(self._samplers) < self._epsilon_greed * self._pop_size:
            sampler = ModelSampler(random_state=self._random_states.getrandbits(32))
        else:
            sampler = ModelSampler(self._model, random_state=self._random_states.getrandbits(32))

        self._samplers.append(sampler)
        return sampler

    def _finish_generation(self, fns):
        # Compute the marginal model of the best pipelines
        indices = best_indices(
            fns, k=int(self._selection * len(fns)), maximize=self._maximize
        )
        samplers: List[ModelSampler] = [self._samplers[i] for i in indices]
        updates: Dict = merge_updates(*[sampler.updates for sampler in samplers])

        # Update the probabilistic model with the marginal model from the best pipelines
        self._model = update_model(self._model, updates, self._learning_factor)
        
        # save an internal state of metaheuristic for other executions
        if self._save == True:
            with open('model-'+self._name+'.pickle', 'wb') as f:
                pickle.dump(self._model, f) 

    def load(self, name_pickle_file):
        """Rewrites the probabilistic distribution of metaheuristic with the value of the name model.
        """
        
        with open(name_pickle_file) as f:
            loaded_obj = pickle.load(f)
        self._model = loaded_obj


from tensorflow.keras.layers import Conv1D as _Conv1D
from tensorflow.keras.layers import Conv2D as _Conv2D
from tensorflow.keras.layers import MaxPooling2D as _MaxPooling2D
from tensorflow.keras.layers import Dense as _Dense
from tensorflow.keras.layers import Embedding as _Embedding
from tensorflow.keras.layers import LSTM as _LSTM
from tensorflow.keras.layers import Bidirectional
from tensorflow.keras.layers import Dropout as _Dropout
from tensorflow.keras.layers import BatchNormalization as _BatchNormalization
from tensorflow.keras.layers import TimeDistributed as _TimeDistributed
from tensorflow.keras.layers import Activation as _Activation
from tensorflow.keras.layers import Flatten as _Flatten
from tensorflow.keras.layers import Reshape as _Reshape
from tensorflow.keras import regularizers

from autogoal.grammar import Boolean, Categorical, Discrete, Continuous
from autogoal.utils import nice_repr


@nice_repr
class Seq2SeqLSTM(_LSTM):
    def __init__(
        self,
        units: Discrete(32, 1024),
        activation: Categorical("tanh", "sigmoid", "relu", "linear"),
        recurrent_activation: Categorical("tanh", "sigmoid", "relu", "linear"),
        dropout: Continuous(0, 0.5),
        recurrent_dropout: Continuous(0, 0.5),
    ):
        super().__init__(
            units=units,
            activation=activation,
            recurrent_activation=recurrent_activation,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            return_sequences=True,
        )


@nice_repr
class Seq2VecLSTM(_LSTM):
    def __init__(
        self,
        units: Discrete(32, 1024),
        activation: Categorical("tanh", "sigmoid", "relu", "linear"),
        recurrent_activation: Categorical("tanh", "sigmoid", "relu", "linear"),
        dropout: Continuous(0, 0.5),
        recurrent_dropout: Continuous(0, 0.5),
    ):
        super().__init__(
            units=units,
            activation=activation,
            recurrent_activation=recurrent_activation,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            return_sequences=False,
        )


@nice_repr
class Seq2SeqBiLSTM(Bidirectional):
    def __init__(
        self,
        merge_mode: Categorical("sum", "mul", "concat", "ave"),
        units: Discrete(32, 1024),
        activation: Categorical("tanh", "sigmoid", "relu", "linear"),
        recurrent_activation: Categorical("tanh", "sigmoid", "relu", "linear"),
        dropout: Continuous(0, 0.5),
        recurrent_dropout: Continuous(0, 0.5),
    ):
        super().__init__(
            layer=_LSTM(
                units=units,
                activation=activation,
                recurrent_activation=recurrent_activation,
                dropout=dropout,
                recurrent_dropout=recurrent_dropout,
                return_sequences=True,
            ),
            merge_mode=merge_mode,
        )


@nice_repr
class Seq2VecBiLSTM(Bidirectional):
    def __init__(
        self,
        merge_mode: Categorical("sum", "mul", "concat", "ave"),
        units: Discrete(32, 1024),
        activation: Categorical("tanh", "sigmoid", "relu", "linear"),
        recurrent_activation: Categorical("tanh", "sigmoid", "relu", "linear"),
        dropout: Continuous(0, 0.5),
        recurrent_dropout: Continuous(0, 0.5),
    ):
        super().__init__(
            layer=_LSTM(
                units=units,
                activation=activation,
                recurrent_activation=recurrent_activation,
                dropout=dropout,
                recurrent_dropout=recurrent_dropout,
                return_sequences=False,
            ),
            merge_mode=merge_mode,
        )


@nice_repr
class Reshape2D(_Reshape):
    def __init__(self):
        super().__init__(target_shape=(-1, 1))


@nice_repr
class Embedding(_Embedding):
    def __init__(self, output_dim: Discrete(32, 128)):
        super().__init__(input_dim=1000, output_dim=output_dim)


@nice_repr
class Dense(_Dense):
    def __init__(self, units: Discrete(128, 1024), **kwargs):
        super().__init__(units=units, **kwargs)


@nice_repr
class Conv1D(_Conv1D):
    def __init__(self, filters: Discrete(2, 8), kernel_size: Categorical(3, 5, 7)):
        super().__init__(
            filters=2 ** filters, kernel_size=kernel_size, padding="causal"
        )


@nice_repr
class Conv2D(_Conv2D):
    def __init__(
        self,
        filters: Discrete(2, 8),
        kernel_size: Categorical(3, 5, 7),
        l1: Continuous(0, 1e-3),
        l2: Continuous(0, 1e-3),
    ):
        self.l1 = l1
        self.l2 = l2
        super().__init__(
            filters=2 ** filters,
            kernel_size=(kernel_size, kernel_size),
            kernel_regularizer=regularizers.l1_l2(l1=l1, l2=l2),
            padding="same",
            data_format="channels_last",
        )


@nice_repr
class MaxPooling2D(_MaxPooling2D):
    def __init__(self):
        super().__init__(
            data_format="channels_last", padding="same",
        )


@nice_repr
class TimeDistributed(_TimeDistributed):
    def __init__(self, layer: Dense):
        super().__init__(layer)


@nice_repr
class Dropout(_Dropout):
    def __init__(self, rate: Continuous(0, 0.5)):
        super().__init__(rate=rate)


@nice_repr
class BatchNormalization(_BatchNormalization):
    def __init__(self):
        super().__init__()


@nice_repr
class Activation(_Activation):
    def __init__(
        self,
        function: Categorical(
            "elu",
            "selu",
            "relu",
            "tanh",
            "sigmoid",
            "hard_sigmoid",
            "exponential",
            "linear",
        ),
    ):
        self.function = function
        super().__init__(activation=function)


@nice_repr
class Flatten(_Flatten):
    def __init__(self):
        super().__init__()



from typing import Optional

import collections
import numpy as np
from tensorflow.keras.callbacks import EarlyStopping, TerminateOnNaN
from tensorflow.keras.layers import Dense, Input, TimeDistributed, concatenate
from tensorflow.keras.models import Model
from tensorflow.keras.utils import to_categorical
from tensorflow.keras.preprocessing.image import (
    ImageDataGenerator as _ImageDataGenerator,
)

from autogoal.contrib.keras._grammars import build_grammar, generate_grammar, Modules
from autogoal.grammar import (
    Graph,
    GraphGrammar,
    Sampler,
    Boolean,
    Categorical,
    Continuous,
    Discrete,
)
from autogoal.kb import (
    CategoricalVector,
    List,
    MatrixContinuousDense,
    Postag,
    Tensor3,
    Tensor4,
    Tuple,
)
from autogoal.utils import nice_repr


@nice_repr
class KerasNeuralNetwork:
    def __init__(
        self,
        grammar: GraphGrammar,
        optimizer: Categorical("sgd", "adam", "rmsprop"),
        epochs=10,
        early_stop=3,
        validation_split=0.1,
        **compile_kwargs,
    ):
        self.optimizer = optimizer
        self._grammar = grammar
        self._epochs = epochs
        self._compile_kwargs = compile_kwargs
        self._model: Optional[Model] = None
        self._mode = "train"
        self._graph = None
        self._validation_split = validation_split
        self._early_stop = early_stop

    def train(self):
        self._mode = "train"

    def eval(self):
        self._mode = "eval"

    def run(self, input):
        X, y = input
        if self._mode == "train":
            self.fit(X, y)
            return y
        if self._mode == "eval":
            return self.predict(X)

        assert False, "Invalid mode %s" % self._mode

    def __repr__(self):
        nodes = len(self._graph.nodes) if self._graph is not None else None
        return f"{self.__class__.__name__}(nodes={nodes}, compile_kwargs={self._compile_kwargs})"

    def __nice_repr_hook__(self, names, values):
        graph = []

        if self._graph is not None:
            for node, _ in self._graph.build_order():
                graph.append(node)

        names.append("graph")
        values.append(graph)

    @property
    def model(self):
        if self._model is None:
            raise TypeError("You need to call `fit` first to generate the model.")

        return self._model

    def sample(self, sampler: Sampler = None, max_iterations=100):
        if sampler is None:
            sampler = Sampler()

        self._graph = self._grammar.sample(
            sampler=sampler, max_iterations=max_iterations
        )
        return self

    def _build_nn(self, graph: Graph, X, y):
        input_x = self._build_input(X)

        def build_model(layer, _, previous_layers):
            if not previous_layers:
                return layer(input_x)

            if len(previous_layers) > 1:
                incoming = concatenate(previous_layers)
            else:
                incoming = previous_layers[0]

            return layer(incoming)

        output_y = graph.apply(build_model) or [input_x]
        final_ouput = self._build_output(output_y, y)

        if "optimizer" not in self._compile_kwargs:
            self._compile_kwargs["optimizer"] = self.optimizer

        self._model = Model(inputs=input_x, outputs=final_ouput)
        self._model.compile(**self._compile_kwargs)

    def _build_input(self, X):
        raise NotImplementedError()

    def _build_output(self, outputs, y):
        return outputs

    def fit(self, X, y, **kwargs):
        if self._graph is None:
            raise TypeError("You must call `sample` to generate the internal model.")

        self._build_nn(self._graph, X, y)
        self.model.summary()
        self._fit_model(X, y, **kwargs)

    def _fit_model(self, X, y, **kwargs):
        self.model.fit(
            x=X,
            y=y,
            epochs=self._epochs,
            callbacks=[
                EarlyStopping(patience=self._early_stop, restore_best_weights=True),
                TerminateOnNaN(),
            ],
            validation_split=self._validation_split,
            **kwargs,
        )

    def predict(self, X):
        return self.model.predict(X)


class KerasClassifier(KerasNeuralNetwork):
    def __init__(
        self, optimizer: Categorical("sgd", "adam", "rmsprop"), grammar=None, **kwargs
    ):
        self._classes = None
        self._num_classes = None
        super().__init__(
            grammar=grammar or self._build_grammar(), optimizer=optimizer, **kwargs
        )

    def _build_grammar(self):
        return build_grammar(features=True)

    def _build_input(self, X):
        return Input(shape=(X.shape[1],))

    def _build_output_layer(self, y):
        self._num_classes = y.shape[1]

        if "loss" not in self._compile_kwargs:
            self._compile_kwargs["loss"] = "categorical_crossentropy"
            self._compile_kwargs["metrics"] = ["accuracy"]

        return Dense(units=self._num_classes, activation="softmax")

    def _build_output(self, outputs, y):
        if len(outputs) > 1:
            outputs = concatenate(outputs)
        else:
            outputs = outputs[0]

        return self._build_output_layer(y)(outputs)

    def fit(self, X, y):
        self._classes = {k: v for k, v in zip(set(y), range(len(y)))}
        self._inverse_classes = {v: k for k, v in self._classes.items()}
        y = [self._classes[yi] for yi in y]
        y = to_categorical(y)
        return super(KerasClassifier, self).fit(X, y)

    def predict(self, X):
        if self._classes is None:
            raise TypeError(
                "You must call `fit` before `predict` to learn class mappings."
            )

        predictions = super(KerasClassifier, self).predict(X)
        predictions = predictions.argmax(axis=-1)
        return [self._inverse_classes[yi] for yi in predictions]

    def run(
        self, input: Tuple(MatrixContinuousDense(), CategoricalVector())
    ) -> CategoricalVector():
        return super().run(input)


@nice_repr
class KerasImagePreprocessor(_ImageDataGenerator):
    """Augment a dataset of images by making changes to the original training set.

    Applies standard dataset augmentation strategies, such as rotating,
    scaling and fliping the image.
    Uses the `ImageDataGenerator` class from keras.

    The parameter `grow_size` determines how many new images will be created for each original image.
    The remaining parameters are passed to `ImageDataGenerator`.
    """

    def __init__(
        self,
        featurewise_center: Boolean(),
        samplewise_center: Boolean(),
        featurewise_std_normalization: Boolean(),
        samplewise_std_normalization: Boolean(),
        rotation_range: Discrete(0, 15),
        width_shift_range: Continuous(0, 0.25),
        height_shift_range: Continuous(0, 0.25),
        shear_range: Continuous(0, 15),
        zoom_range: Continuous(0, 0.25),
        horizontal_flip: Boolean(),
        vertical_flip: Boolean(),
    ):
        super().__init__(
            featurewise_center=featurewise_center,
            samplewise_center=samplewise_center,
            featurewise_std_normalization=featurewise_std_normalization,
            samplewise_std_normalization=samplewise_std_normalization,
            rotation_range=rotation_range,
            width_shift_range=width_shift_range,
            height_shift_range=height_shift_range,
            shear_range=shear_range,
            zoom_range=zoom_range,
            horizontal_flip=horizontal_flip,
            vertical_flip=vertical_flip,
        )


class KerasImageClassifier(KerasClassifier):
    def __init__(
        self,
        preprocessor: KerasImagePreprocessor,
        optimizer: Categorical("sgd", "adam", "rmsprop"),
        **kwargs,
    ):
        self.preprocessor = preprocessor
        super().__init__(optimizer=optimizer, **kwargs)

    def _build_grammar(self):
        return generate_grammar(
            Modules.Preprocessing.Conv2D(),
            # Modules.Features.Dense()
        )

    def _fit_model(self, X, y, **kwargs):
        self.preprocessor.fit(X)
        batch_size = 64
        validation_size = min(int(0.25 * len(X)), 10 * batch_size)

        Xtrain, Xvalid = X[:-validation_size], X[-validation_size:]
        ytrain, yvalid = y[:-validation_size], y[-validation_size:]

        self.model.fit(
            self.preprocessor.flow(Xtrain, ytrain, batch_size=batch_size),
            steps_per_epoch=len(Xtrain) // batch_size,
            epochs=self._epochs,
            callbacks=[
                EarlyStopping(patience=self._early_stop, restore_best_weights=True),
                TerminateOnNaN(),
            ],
            validation_data=(Xvalid, yvalid),
            **kwargs,
        )

    def run(self, input: Tuple(Tensor4(), CategoricalVector())) -> CategoricalVector():
        return super().run(input)

    def _build_input(self, X):
        return Input(shape=X.shape[1:])


class KerasSequenceClassifier(KerasClassifier):
    def _build_grammar(self):
        return build_grammar(preprocessing=True, reduction=True, features=True)

    def _build_input(self, X):
        return Input(shape=(None, X.shape[2]))

    def run(self, input: Tuple(Tensor3(), CategoricalVector())) -> CategoricalVector():
        return super().run(input)


# !!! warning
# These imports are taken from <https://github.com/tensorflow/addons/pull/377>
# since the CRF layer has not yet landed in tensorflow
# TODO: Make sure to replace this when tensorflow merges this commit
from autogoal.contrib.keras._crf import CRF
from autogoal.contrib.keras._crf import crf_loss


class KerasSequenceTagger(KerasNeuralNetwork):
    def __init__(
        self,
        decode: Categorical("dense", "crf"),
        optimizer: Categorical("sgd", "adam", "rmsprop"),
        grammar=None,
        **kwargs,
    ):
        self._classes = None
        self._num_classes = None

        if decode not in ["dense", "crf"]:
            raise ValueError(f"Invalid decode={decode}")

        self.decode = decode
        super().__init__(
            grammar=grammar or self._build_grammar(), optimizer=optimizer, **kwargs
        )

    def _build_grammar(self):
        from autogoal.grammar._graph import Epsilon

        return GraphGrammar(Epsilon())

    def _build_input(self, X):
        return Input(shape=(None, X[0].shape[-1]))

    def _build_output(self, outputs, y):
        if "loss" not in self._compile_kwargs:
            self._compile_kwargs["metrics"] = ["accuracy"]

            if self.decode == "dense":
                self._compile_kwargs["loss"] = "categorical_crossentropy"
            elif self.decode == "crf":
                self._compile_kwargs["loss"] = crf_loss

        if len(outputs) > 1:
            outputs = concatenate(outputs)
        else:
            outputs = outputs[0]

        if self.decode == "dense":
            dense = Dense(units=len(self._classes), activation="softmax")
            return TimeDistributed(dense)(outputs)
        elif self.decode == "crf":
            crf = CRF(units=len(self._classes))
            return crf(outputs)

    def fit(self, X, y):
        distinct_classes = set(x for yi in y for x in yi)

        self._classes = {
            k: v for k, v in zip(distinct_classes, range(len(distinct_classes)))
        }
        self._inverse_classes = {v: k for k, v in self._classes.items()}

        y = [[self._classes[x] for x in yi] for yi in y]
        return super().fit(X, y)

    def _encode(self, xi, yi):
        if self.decode == "dense":
            yi = to_categorical(yi, len(self._classes))
            
        return (
            np.expand_dims(xi, axis=0),
            np.expand_dims(yi, axis=0),
        )
        
    def _generate_batches(self, X, y):
        while True:
            for xi, yi in zip(X, y):
                yield self._encode(xi, yi)

    def _fit_model(self, X, y, **kwargs):
        self.model.fit(
            self._generate_batches(X, y),
            steps_per_epoch=len(X),
            epochs=self._epochs,
            callbacks=[
                EarlyStopping(
                    patience=self._early_stop,
                    restore_best_weights=True,
                    monitor="accuracy",
                ),
                TerminateOnNaN(),
            ],
            **kwargs,
        )

    def _decode(self, predictions):
        if self.decode == "dense":
            predictions = [pr.argmax(axis=-1) for pr in predictions]
            return [[self._inverse_classes[x] for x in yi[0]] for yi in predictions]
        elif self.decode == "crf":
            return [[self._inverse_classes[x] for x in yi[0]] for yi in predictions]

    def predict(self, X):
        if self._classes is None:
            raise TypeError(
                "You must call `fit` before `predict` to learn class mappings."
            )

        predictions = [self.model.predict(np.expand_dims(xi, axis=0)) for xi in X]
        return self._decode(predictions)

    def run(
        self, input: Tuple(List(MatrixContinuousDense()), List(List(Postag())))
    ) -> List(List(Postag())):
        return super().run(input)



import inspect
import typer
import collections

from pathlib import Path
from autogoal import kb
from autogoal.contrib import find_classes


app = typer.Typer(name="AutoGOAL")
contrib_app = typer.Typer(name="contrib")

app.add_typer(contrib_app)


@app.callback()
def main():
    """
    Manage AutoGOAL directly from the CLI.
    """


@app.command()
def demo():
    """
    Launch streamlit demo.
    """

    try:
        from streamlit.bootstrap import run

        run(Path(__file__).parent / "contrib" / "streamlit" / "demo.py", "", "")
    except ImportError:
        print("(!) Too run the demo you need streamlit installed.")
        print("(!) Fix it by running `pip install autogoal[streamlit]`.")


@contrib_app.callback()
def contrib_main():
    """
    Inspect contrib libraries and algorithms.
    """


@contrib_app.command("list")
def contrib_list(
    verbose: bool = False,
    include: str = None,
    exclude: str = None,
    input: str = None,
    output: str = None,
):
    """
    List all currently available contrib algorithms.
    """
    from autogoal.contrib import find_classes

    classes = find_classes(include=include, exclude=exclude, input=input, output=output)
    classes_by_contrib = collections.defaultdict(list)
    max_cls_name_length = 0

    for cls in classes:
        max_cls_name_length = max(max_cls_name_length, len(cls.__name__))
        classes_by_contrib[str(cls).split(".")[2]].append(cls)

    typer.echo(
        f"⚙️  Found a total of {len(classes)} matching algorithms.", color="blue"
    )

    for contrib, clss in classes_by_contrib.items():
        typer.echo(f"🛠️  {contrib}: {len(clss)} algorithms.")

        if verbose:
            for cls in clss:
                sig = inspect.signature(cls.run)
                typer.echo(f" 🔹 {cls.__name__.ljust(max_cls_name_length)} : {sig.parameters['input'].annotation} -> {sig.return_annotation}")


if __name__ == "__main__":
    app()



# coding: utf-8

import numpy as np
import os

from autogoal.datasets import download, datapath

def load(white=True, red=True, max_examples=None):
    if not red and not white:
        raise ValueError("Either red or white must be selected")

    download("wine_quality")

    f_white = open(datapath("wine_quality") / "winequality-white.csv", "r")
    f_red = open(datapath("wine_quality") / "winequality-red.csv", "r")

    X = []
    y = []

    if white:
        title_line = True
        for i in f_white.readlines():

            if max_examples and len(X) >= max_examples:
                break

            if title_line == True:
                title_line = False
                continue

            clean_line = i.strip().split(";")

            X.append([1,0]+[float(i) for i in clean_line[:-1]])
            y.append(float(clean_line[-1]))


    if red:
        title_line = True
        for i in f_red.readlines():

            if max_examples and len(X) >= max_examples:
                break

            if title_line == True:
                title_line = False
                continue

            clean_line = i.strip().split(";")

            X.append([0,1]+[float(i) for i in clean_line[:-1]])
            y.append(float(clean_line[-1]))


    return np.asarray(X), np.asarray(y)
